# Copyright European Organization for Nuclear Research (CERN) since 2012
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
   Hermes is a daemon that get the messages and sends them to external services (influxDB, ES, ActiveMQ).
"""

import calendar
import datetime
import functools
import json
import logging
import re
import smtplib
import sys
import threading
import time
from configparser import NoOptionError, NoSectionError
from email.mime.text import MIMEText
from typing import TYPE_CHECKING, Any, Optional, Union

import requests
from requests.auth import HTTPBasicAuth

import rucio.db.sqla.util
from rucio.common.config import config_get, config_get_bool, config_get_list
from rucio.common.exception import DatabaseException
from rucio.common.logging import formatted_logger, setup_logging
from rucio.common.stomp_utils import ListenerBase, StompConnectionManager
from rucio.core.message import delete_messages, retrieve_messages
from rucio.core.monitor import MetricManager
from rucio.daemons.common import run_daemon

if TYPE_CHECKING:
    from collections.abc import Iterable
    from types import FrameType

    from rucio.common.types import LoggerFunction
    from rucio.daemons.common import HeartbeatHandler

logging.getLogger("requests").setLevel(logging.CRITICAL)

METRICS = MetricManager(module=__name__)
graceful_stop = threading.Event()
DAEMON_NAME = "hermes"

RECONNECT_COUNTER = METRICS.counter(
    name="reconnect.{host}",
    documentation="Counts Hermes reconnects to different ActiveMQ brokers",
    labelnames=("host",),
)


def default(datetype: Union[datetime.date, datetime.datetime]) -> str:
    if isinstance(datetype, (datetime.date, datetime.datetime)):
        return datetype.isoformat()


class HermesListener(ListenerBase):
    """
    Hermes Listener
    """


def deliver_emails(messages: "Iterable[dict[str, Any]]", logger: "LoggerFunction") -> list[int]:
    """
    Sends emails

    :param messages:           The list of messages.
    :param logger:             The logger object.

    :returns:                  List of message_id to delete
    """

    email_from = config_get("messaging-hermes", "email_from")
    send_email = config_get_bool(
        "messaging-hermes", "send_email", raise_exception=False, default=True
    )
    to_delete = []
    for message in messages:
        if message['event_type'] == 'email':
            msg = MIMEText(message['payload']['body'])
            msg['From'] = email_from
            msg['To'] = ', '.join(message['payload']['to'])
            msg['Subject'] = message['payload']['subject']

            try:
                if send_email:
                    smtp = smtplib.SMTP()
                    smtp.connect()
                    smtp.sendmail(
                        msg["From"], message["payload"]["to"], msg.as_string()
                    )
                    smtp.quit()
                to_delete.append(message["id"])
            except Exception as error:
                logger(logging.ERROR, "Cannot send email : %s", str(error))
        else:
            to_delete.append(message["id"])
            continue
    return to_delete


def submit_to_elastic(
        messages: "Iterable[dict[str, Any]]",
        endpoint: str,
        logger: "LoggerFunction"
) -> int:
    """
    Aggregate a list of message to ElasticSearch

    :param messages:           The list of messages.
    :param endpoint:           The ES endpoint were to send the messages.
    :param logger:             The logger object.

    :returns:                  HTTP status code. 200 and 204 OK. Rest is failure.
    """
    text = ""
    elastic_username = config_get("hermes", "elastic_username",
                                  raise_exception=False, default=None)
    elastic_password = config_get("hermes", "elastic_password",
                                  raise_exception=False, default=None)
    auth = None
    if elastic_username and elastic_password:
        auth = HTTPBasicAuth(elastic_username, elastic_password)

    for message in messages:
        text += '{ "index":{ } }\n%s\n' % json.dumps(message, default=default)
    res = requests.post(endpoint,
                        data=text,
                        headers={"Content-Type": "application/json"},
                        auth=auth)
    return res.status_code


def aggregate_to_influx(
    messages: "Iterable[dict[str, Any]]",
    bin_size: str,
    endpoint: str,
    logger: "LoggerFunction"
) -> int:
    """
    Aggregate a list of message using a certain bin_size
    and submit them to a InfluxDB endpoint

    :param messages:           The list of messages.
    :param bin_size:           The size of the bins for the aggregation (e.g. 10m, 1h, etc.).
    :param endpoint:           The InfluxDB endpoint were to send the messages.
    :param logger:             The logger object.

    :returns:                  HTTP status code. 200 and 204 OK. Rest is failure.
    """
    bins = {}
    dtime = datetime.datetime.now()
    microsecond = dtime.microsecond

    for message in messages:
        event_type = message["event_type"]
        payload = message["payload"]
        if event_type in ["transfer-failed", "transfer-done"]:
            if not payload["transferred_at"]:
                logger(logging.WARNING,
                       "No transferred_at for message. Reason : %s",
                       payload["reason"])
                continue
            transferred_at = time.strptime(
                payload["transferred_at"], "%Y-%m-%d %H:%M:%S"
            )
            if bin_size == "1m":
                transferred_at = int(calendar.timegm(transferred_at)) * 1000000000
                transferred_at += microsecond
            if transferred_at not in bins:
                bins[transferred_at] = {}
            src_rse, dest_rse, activity = (
                payload["src-rse"],
                payload["dst-rse"],
                payload["activity"],
            )
            activity = re.sub(" ", r"\ ", activity)
            key = f"transfer,activity={activity!s},src_rse={src_rse!s},dst_rse={dest_rse!s}"
            if key not in bins[transferred_at]:
                bins[transferred_at][key] = [0, 0, 0, 0]
            if event_type == "transfer-done":
                bins[transferred_at][key][0] += 1
                bins[transferred_at][key][1] += payload["bytes"]
            if event_type == "transfer-failed":
                bins[transferred_at][key][2] += 1
                bins[transferred_at][key][3] += payload["bytes"]
        elif event_type in ["deletion-failed", "deletion-done"]:
            created_at = message["created_at"]
            if bin_size == "1m":
                created_at = created_at.replace(
                    second=0, microsecond=0, tzinfo=datetime.timezone.utc
                ).timestamp()
            created_at = int(created_at) * 1000000000
            created_at += microsecond
            if created_at not in bins:
                bins[created_at] = {}
            rse = payload["rse"]
            key = f"deletion,rse={rse!s}"
            if key not in bins[created_at]:
                bins[created_at][key] = [0, 0, 0, 0]
            if event_type == "deletion-done":
                bins[created_at][key][0] += 1
                bins[created_at][key][1] += payload["bytes"]
            if event_type == "deletion-failed":
                bins[created_at][key][2] += 1
                bins[created_at][key][3] += payload["bytes"]

    points = ""
    for timestamp, entries in bins.items():
        for key, metrics in entries.items():
            event_type = key.split(",")[0]
            points += (f"{key!s} "
                       f"nb_{event_type!s}_done={metrics[0]!s},"
                       f"bytes_{event_type!s}_done={metrics[1]!s},"
                       f"nb_{event_type!s}_failed={metrics[2]!s},"
                       f"bytes_{event_type!s}_failed={metrics[3]!s} "
                       rf"{timestamp!s}\n")

    influx_token = config_get("hermes", "influxdb_token", False, None)
    headers = {}
    if influx_token:
        headers["Authorization"] = f"Token {influx_token!s}"
    if points:
        res = requests.post(endpoint, headers=headers, data=points)
        logger(logging.DEBUG, "%s", str(res.text))
        return res.status_code
    return 204


def build_message_dict(
        bulk: int,
        thread: int,
        total_threads: int,
        message_dict: dict[str, list[dict[str, Any]]],
        logger: "LoggerFunction",
        service: Optional[str] = None,
) -> None:
    """
    Retrieves messages from the database and builds a dictionary with the keys being the services, and the values a list of the messages (built up of dictionary / json information)

    :param bulk:               Integer for number of messages to retrieve.
    :param thread:             Passed to thread in retrieve_messages for Identifier of the caller thread as an integer.
    :param total_threads:      Passed to total_threads for Maximum number of threads as an integer.
    :param message_dict:       Either empty dictionary to be built, or build upon when using query_by_service.
    :param logger:             The logger object.
    :param service:            When passed, only returns messages table for this specific service.

    :returns:                  None, but builds on the dictionary message_dict passed to this fuction (for when querying multiple services).
    """
    start_time = time.time()
    messages = retrieve_messages(
        bulk=bulk,
        old_mode=False,
        thread=thread,
        total_threads=total_threads,
        service_filter=service,
    )

    if messages:
        if service is not None:
            # query_by_service dictionary build behaviour
            message_dict[service] = messages.copy()
            logger(
                logging.DEBUG,
                "Retrieved %i messages retrieved in %s seconds for %s service.",
                len(messages),
                time.time() - start_time,
                service,
            )

        else:
            # default dictionary build behaviour
            for message in messages:
                service = message["services"]
                if service is not None:
                    if service not in message_dict:
                        message_dict[service] = []
                    message_dict[service].append(message)
        logger(
            logging.DEBUG,
            "Retrieved %i messages retrieved in %s seconds",
            len(messages),
            time.time() - start_time,
        )
    else:
        logger(
            logging.INFO,
            "No messages retrieved in %s seconds",
            time.time() - start_time,
        )


def hermes(once: bool = False, bulk: int = 1000, sleep_time: int = 10) -> None:
    """
    Creates a Hermes Worker that can submit messages to different services (InfluXDB, ElasticSearch, ActiveMQ)
    The list of services need to be define in the config service in the hermes section.
    The list of endpoints need to be defined in rucio.cfg in the hermes section.

    :param once:       Run only once.
    :param bulk:       The number of requests to process.
    :param sleep_time: Time between two cycles.
    """
    run_daemon(
        once=once,
        graceful_stop=graceful_stop,
        executable=DAEMON_NAME,
        partition_wait_time=1,
        sleep_time=sleep_time,
        run_once_fnc=functools.partial(
            run_once,
            bulk=bulk,
        ),
    )


def run_once(heartbeat_handler: "HeartbeatHandler", bulk: int, **_kwargs) -> bool:

    worker_number, total_workers, logger = heartbeat_handler.live()
    try:
        services_list = config_get_list("hermes", "services_list")
    except (NoOptionError, NoSectionError, RuntimeError):
        logger(logging.DEBUG, "No services found, exiting")
        sys.exit(1)

    if "influx" in services_list:
        influx_endpoint = None
        try:
            influx_endpoint = config_get("hermes", "influxdb_endpoint", False, None)
            if not influx_endpoint:
                logger(
                    logging.ERROR,
                    "InfluxDB defined in the services list, but no endpoint can be found",
                )
        except Exception as err:
            logger(logging.ERROR, str(err))
    if "elastic" in services_list:
        elastic_endpoint = None
        try:
            elastic_endpoint = config_get("hermes", "elastic_endpoint", False, None)
            if not elastic_endpoint:
                logger(
                    logging.ERROR,
                    "Elastic defined in the services list, but no endpoint can be found",
                )
        except Exception as err:
            logger(logging.ERROR, str(err))
    conns = None
    if "activemq" in services_list:
        conn_mgr = StompConnectionManager(config_section='messaging-hermes', logger=logger)
        conn_mgr.set_listener_factory("rucio-hermes", HermesListener, heartbeats=conn_mgr.config.heartbeats)

    worker_number, total_workers, logger = heartbeat_handler.live()
    message_dict = {}
    query_by_service = config_get_bool("hermes", "query_by_service", default=False)

    # query_by_service is a toggleable behaviour switch between collecting bulk number of messages across all services when false, to collecting bulk messages from each service when true.
    if query_by_service:
        for service in services_list:
            build_message_dict(
                bulk=bulk,
                thread=worker_number,
                total_threads=total_workers,
                message_dict=message_dict,
                logger=logger,
                service=service,
            )
    else:
        build_message_dict(
            bulk=bulk,
            thread=worker_number,
            total_threads=total_workers,
            message_dict=message_dict,
            logger=logger
        )

    if message_dict:
        to_delete = []

        if "influx" in message_dict and influx_endpoint:
            # For influxDB, bulk submission, either everything succeeds or fails
            t_time = time.time()
            logger(logging.DEBUG, "Will submit to influxDB")
            try:
                state = aggregate_to_influx(
                    messages=message_dict["influx"],
                    bin_size="1m",
                    endpoint=influx_endpoint,
                    logger=logger,
                )
                if state in [204, 200]:
                    logger(
                        logging.INFO,
                        "%s messages successfully submitted to influxDB in %s seconds",
                        len(message_dict["influx"]),
                        time.time() - t_time,
                    )
                    for message in message_dict["influx"]:
                        to_delete.append(message)
                else:
                    logger(
                        logging.ERROR,
                        "Failure to submit %s messages to influxDB. Returned status: %s",
                        len(message_dict["influx"]),
                        state,
                    )
            except Exception as error:
                logger(logging.ERROR, "Error sending to InfluxDB : %s", str(error))

        if "elastic" in message_dict and elastic_endpoint:
            # For elastic, bulk submission, either everything succeeds or fails
            t_time = time.time()
            try:
                state = submit_to_elastic(
                    messages=message_dict["elastic"],
                    endpoint=elastic_endpoint,
                    logger=logger,
                )
                if state in [200, 204]:
                    logger(
                        logging.INFO,
                        "%s messages successfully submitted to elastic in %s seconds",
                        len(message_dict["elastic"]),
                        time.time() - t_time,
                    )
                    for message in message_dict["elastic"]:
                        to_delete.append(message)
                else:
                    logger(
                        logging.ERROR,
                        "Failure to submit %s messages to elastic. Returned status: %s",
                        len(message_dict["elastic"]),
                        state,
                    )
            except Exception as error:
                logger(logging.ERROR, "Error sending to Elastic : %s", str(error))

        if "email" in message_dict:
            t_time = time.time()
            try:
                messages_sent = deliver_emails(
                    messages=message_dict["email"], logger=logger
                )
                logger(
                    logging.INFO,
                    "%s messages successfully submitted by emails in %s seconds",
                    len(message_dict["email"]),
                    time.time() - t_time,
                )
                for message in message_dict["email"]:
                    if message["id"] in messages_sent:
                        to_delete.append(message)
            except Exception as error:
                logger(logging.ERROR, "Error sending email : %s", str(error))

        if "activemq" in message_dict and conns:
            t_time = time.time()
            try:
                messages_sent = conn_mgr.deliver_messages(messages=message_dict["activemq"])
                logger(
                    logging.INFO,
                    "%s messages successfully submitted to ActiveMQ in %s seconds",
                    len(message_dict["activemq"]),
                    time.time() - t_time,
                )
                for message in message_dict["activemq"]:
                    if message["id"] in messages_sent:
                        to_delete.append(message)
            except Exception as error:
                logger(logging.ERROR, "Error sending to ActiveMQ : %s", str(error))

    logger(logging.INFO, "Deleting %s messages", len(to_delete))
    to_delete = [
        {
            "id": message["id"],
            "created_at": message["created_at"],
            "updated_at": message["created_at"],
            "payload": str(message["payload"]),
            "event_type": message["event_type"],
            "services": message["services"]
        }
        for message in to_delete
    ]
    delete_messages(messages=to_delete)
    must_sleep = True
    return must_sleep


def stop(signum: Optional[int] = None, frame: Optional["FrameType"] = None) -> None:
    """
    Graceful exit.
    """
    logging.info("Caught CTRL-C - waiting for cycle to end before shutting down")
    graceful_stop.set()


def run(
    once: bool = False,
    threads: int = 1,
    bulk: int = 1000,
    sleep_time: int = 10,
    broker_timeout: int = 3,
) -> None:
    """
    Starts up the hermes threads.
    """
    setup_logging(process_name=DAEMON_NAME)
    logger = formatted_logger(logging.log, DAEMON_NAME + ' %s')

    if rucio.db.sqla.util.is_old_db():
        raise DatabaseException("Database was not updated, daemon won't start")

    logger(logging.INFO, "starting hermes threads")
    thread_list = []
    for _ in range(threads):
        her_thread = threading.Thread(target=hermes, kwargs={"once": once, "bulk": bulk, "sleep_time": sleep_time})
        her_thread.start()
        thread_list.append(her_thread)

    logger(logging.DEBUG, thread_list)
    while [thread.join(timeout=3.14) for thread in thread_list if thread.is_alive()]:
        pass



# Copyright European Organization for Nuclear Research (CERN) since 2012
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import base64
import logging
import os.path
import time
from urllib.parse import parse_qs, urlencode, urlparse

from rucio.common import exception
from rucio.common.bittorrent import construct_torrent
from rucio.common.extra import import_extras
from rucio.common.utils import resolve_ip
from rucio.rse import rsemanager
from rucio.rse.protocols.protocol import RSEProtocol

EXTRA_MODULES = import_extras(['libtorrent'])

lt = None
if EXTRA_MODULES['libtorrent']:
    import libtorrent as lt  # pylint: disable=E0401

if getattr(rsemanager, 'CLIENT_MODE', None):
    from rucio.client.didclient import DIDClient

    def _fetch_meta_client(rse_id: str, scope: str, name: str):
        return DIDClient().get_metadata(scope=scope, name=name, plugin='all')

    _fetch_meta = _fetch_meta_client
else:
    from rucio.common.types import InternalScope
    from rucio.core.did import get_metadata
    from rucio.core.rse import get_rse_vo

    def _fetch_meta_server(rse_id: str, scope: str, name: str):
        vo = get_rse_vo(rse_id)
        return get_metadata(scope=InternalScope(scope, vo=vo), name=name, plugin='all')

    _fetch_meta = _fetch_meta_server


class Default(RSEProtocol):

    def __init__(self, protocol_attr, rse_settings, logger=logging.log):
        super(Default, self).__init__(protocol_attr, rse_settings, logger=logger)
        self.logger = logger

    def lfns2pfns(self, lfns):
        pfns = {}
        prefix = self.attributes['prefix']

        if not prefix.startswith('/'):
            prefix = ''.join(['/', prefix])
        if not prefix.endswith('/'):
            prefix = ''.join([prefix, '/'])

        host_port = '%s:%s' % (self.attributes['hostname'], str(self.attributes['port']))

        lfns = [lfns] if isinstance(lfns, dict) else lfns
        for lfn in lfns:
            scope, name = lfn['scope'], lfn['name']

            if 'path' in lfn and lfn['path'] is not None:
                path = lfn['path'] if not lfn['path'].startswith('/') else lfn['path'][1:]
            else:
                path = self._get_path(scope=scope, name=name)

            scope_name = '%s:%s' % (scope, name)

            query = {
                'x.pe': host_port,
                'x.rucio_scope': scope,
                'x.rucio_name': name,
                'x.rucio_path': ''.join((prefix, path))
            }
            pfns[scope_name] = 'magnet:?' + urlencode(query)

        return pfns

    def parse_pfns(self, pfns):
        ret = dict()
        pfns = [pfns] if isinstance(pfns, str) else pfns

        for pfn in pfns:
            parsed = urlparse(pfn)
            scheme = parsed.scheme

            query = parse_qs(parsed.query)
            host_port = next(iter(query.get('x.pe', [])), ':')
            hostname, port = host_port.split(':')
            port = int(port)
            path = next(iter(query.get('x.rucio_path', [])), '')
            scope = next(iter(query.get('x.rucio_scope', [])), '')
            name = next(iter(query.get('x.rucio_name', [])), '')

            # Protect against 'lazy' defined prefixes for RSEs in the repository
            if not self.attributes['prefix'].startswith('/'):
                self.attributes['prefix'] = '/' + self.attributes['prefix']
            if not self.attributes['prefix'].endswith('/'):
                self.attributes['prefix'] += '/'

            if self.attributes['hostname'] != hostname:
                if self.attributes['hostname'] != 'localhost':  # In the database empty hostnames are replaced with localhost but for some URIs (e.g. file) a hostname is not included
                    raise exception.RSEFileNameNotSupported('Invalid hostname: provided \'%s\', expected \'%s\'' % (hostname, self.attributes['hostname']))

            if self.attributes['port'] != port:
                raise exception.RSEFileNameNotSupported('Invalid port: provided \'%s\', expected \'%s\'' % (port, self.attributes['port']))

            if not path.startswith(self.attributes['prefix']):
                raise exception.RSEFileNameNotSupported('Invalid prefix: provided \'%s\', expected \'%s\'' % ('/'.join(path.split('/')[0:len(self.attributes['prefix'].split('/')) - 1]),
                                                                                                              self.attributes['prefix']))  # len(...)-1 due to the leading '/

            # Splitting parsed.path into prefix, path, filename
            prefix = self.attributes['prefix']
            path = path.partition(self.attributes['prefix'])[2]
            path = '/'.join(path.split('/')[:-1])
            if not path.startswith('/'):
                path = '/' + path
            if path != '/' and not path.endswith('/'):
                path = path + '/'
            ret[pfn] = {'path': path, 'scope': scope, 'name': name, 'scheme': scheme, 'prefix': prefix, 'port': port, 'hostname': hostname, }

        return ret

    def connect(self):
        pass

    def close(self):
        pass

    def get(self, path, dest, transfer_timeout=None):
        if not lt:
            raise exception.RucioException('The libtorrent python package is required to perform this operation')

        [lfn] = self.parse_pfns([path]).values()
        scope = lfn['scope']
        name = lfn['name']
        hostname = lfn['hostname']
        port = lfn['port']

        meta = _fetch_meta(rse_id=self.rse['id'], scope=scope, name=name)
        pieces_root = base64.b64decode(meta.get('bittorrent_pieces_root', ''))
        if not pieces_root:
            raise exception.RucioException('Torrent metadata missing. Cannot download file.')

        length = meta.get('bytes')
        piece_length = meta.get('bittorrent_piece_length', 0)
        pieces_layers = base64.b64decode(meta.get('bittorrent_pieces_layers', ''))

        _, torrent = construct_torrent(
            scope=scope,
            name=name,
            length=length,
            piece_length=piece_length,
            pieces_root=pieces_root,
            pieces_layers=pieces_layers,
        )

        ses = lt.session()  # type: ignore # noqa
        params = {
            'ti': lt.torrent_info(torrent),  # type: ignore # noqa
            'save_path': os.path.dirname(dest),
            'name': os.path.basename(dest),
            'renamed_files': {0: os.path.basename(dest)},
        }

        handle = ses.add_torrent(params)
        try:
            handle.resume()
            handle.connect_peer((resolve_ip(hostname), port))
            while handle.status().progress != 1.0:
                time.sleep(0.25)
        finally:
            ses.remove_torrent(handle)

    def delete(self, path):
        raise NotImplementedError

    def put(self, source, target, source_dir, transfer_timeout=None):
        raise NotImplementedError

    def rename(self, path, new_path):
        raise NotImplementedError



# Copyright European Organization for Nuclear Research (CERN) since 2012
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Conveyor is a daemon to manage file transfers.
"""
import json
import logging
import threading
import time
import traceback
from typing import TYPE_CHECKING, Any

import rucio.db.sqla.util
from rucio.common import exception
from rucio.common.logging import formatted_logger, setup_logging
from rucio.common.policy import get_policy
from rucio.common.stomp_utils import Connection, ListenerBase, StompConnectionManager
from rucio.core import request as request_core
from rucio.core import transfer as transfer_core
from rucio.core.monitor import MetricManager
from rucio.daemons.common import HeartbeatHandler
from rucio.db.sqla.session import transactional_session
from rucio.transfertool.fts3 import FTS3CompletionMessageTransferStatusReport

if TYPE_CHECKING:
    from types import FrameType

    from sqlalchemy.orm import Session
    from stomp.utils import Frame

    from rucio.common.types import LoggerFunction

logging.getLogger("stomp").setLevel(logging.CRITICAL)

METRICS = MetricManager(module=__name__)
GRACEFUL_STOP = threading.Event()
DAEMON_NAME = 'conveyor-receiver'


class Receiver(ListenerBase):

    def __init__(self,
                 conn: Connection,
                 id_: str,
                 total_threads: int,
                 transfer_stats_manager: request_core.TransferStatsManager,
                 all_vos: bool = False,
                 logger: "LoggerFunction" = logging.log,
                 **kwargs: dict) -> None:
        super().__init__(conn, logger, **kwargs)
        self.__all_vos = all_vos
        self.__id = id_
        self.__total_threads = total_threads
        self._transfer_stats_manager = transfer_stats_manager

    @METRICS.count_it
    def on_message(self, frame: "Frame") -> None:
        msg = json.loads(frame.body)  # type: ignore

        if not self.__all_vos:
            if 'vo' not in msg or msg['vo'] != get_policy():
                return

        if 'job_metadata' in msg.keys() \
           and isinstance(msg['job_metadata'], dict) \
           and 'issuer' in msg['job_metadata'].keys() \
           and str(msg['job_metadata']['issuer']) == 'rucio':

            if ('job_state' in msg.keys() and (str(msg['job_state']) != 'ACTIVE'
                                               or msg.get('job_multihop', False) is True)):
                METRICS.counter('message_rucio').inc()

                self._perform_request_update(msg)

    @transactional_session
    def _perform_request_update(
        self,
        msg: dict[str, Any],
        *,
        session: "Session | None",
        logger: "LoggerFunction" = logging.log
    ) -> None:
        external_host = msg.get('endpnt', None)
        request_id = msg['file_metadata'].get('request_id', None)
        try:
            tt_status_report = FTS3CompletionMessageTransferStatusReport(external_host,
                                                                         request_id=request_id,
                                                                         fts_message=msg)
            if tt_status_report.get_db_fields_to_update(session=session, logger=logger):  # type: ignore
                logging.info('RECEIVED %s', tt_status_report)

                ret = transfer_core.update_transfer_state(
                    tt_status_report=tt_status_report,
                    stats_manager=self._transfer_stats_manager,
                    session=session,
                    logger=logger,
                )
                if ret:
                    METRICS.counter('update_request_state.{updated}').labels(updated=True).inc(delta=ret)
                else:
                    METRICS.counter('update_request_state.{updated}').labels(updated=False).inc()
        except Exception:
            logging.critical(traceback.format_exc())


def receiver(id_: str,
             total_threads: int = 1,
             all_vos: bool = False,
             logger: "LoggerFunction" = logging.log):
    """
    Main loop to consume messages from the FTS3 producer.
    """
    logger(logging.INFO, 'receiver starting')

    conn_mgr = StompConnectionManager(config_section='messaging-fts3', logger=logger)

    logger(logging.INFO, 'receiver started')

    with (HeartbeatHandler(executable=DAEMON_NAME, renewal_interval=30),
          request_core.TransferStatsManager() as transfer_stats_manager):

        conn_mgr.set_listener_factory('rucio-messaging-fts3', Receiver,
                                      id_=id_,
                                      total_threads=total_threads,
                                      transfer_stats_manager=transfer_stats_manager,
                                      all_vos=all_vos,
                                      heartbeats=conn_mgr.config.heartbeats)

        while not GRACEFUL_STOP.is_set():

            conn_mgr.subscribe(id_='rucio-messaging-fts3', ack='auto')

            time.sleep(1)

        conn_mgr.disconnect()


def stop(signum: "int | None" = None, frame: "FrameType | None" = None) -> None:
    """
    Graceful exit.
    """

    GRACEFUL_STOP.set()


def run(once: bool = False, total_threads: int = 1) -> None:
    """
    Starts up the receiver thread
    """
    setup_logging(process_name=DAEMON_NAME)
    logger = formatted_logger(logging.log, DAEMON_NAME + ' %s')

    if rucio.db.sqla.util.is_old_db():
        raise exception.DatabaseException('Database was not updated, daemon won\'t start')

    logger(logging.INFO, 'starting receiver thread')
    threads = []
    for i in range(total_threads):
        rec_thread = threading.Thread(target=receiver,
                                      kwargs={'id_': i, 'logger': logger, 'total_threads': total_threads})
        rec_thread.start()
        threads.append(rec_thread)

    logger(logging.INFO, 'waiting for interrupts')

    while [thread.join(timeout=3.14) for thread in threads if thread.is_alive()]:
        pass



# Copyright European Organization for Nuclear Research (CERN) since 2012
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import click

from rucio.cli.bin_legacy.rucio import list_account_usage
from rucio.cli.bin_legacy.rucio_admin import (
    add_account,
    add_account_attribute,
    ban_account,
    delete_account,
    delete_account_attribute,
    delete_limits,
    identity_delete,
    info_account,
    list_account_attributes,
    list_accounts,
    list_identities,
    set_limits,
    unban_account,
    update_account,
)
from rucio.cli.bin_legacy.rucio_admin import identity_add as legacy_identity_add
from rucio.cli.utils import Arguments


@click.group()
def account():
    """Methods to add or change accounts for users, groups, and services. Used to assign privileges"""


@account.command("add")
@click.argument("account-name")
@click.argument("account-type", type=click.Choice(["USER", "GROUP", "SERVICE"]))
@click.option("--email", type=str, help="Email address associated with the account")
@click.pass_context
def add_(ctx, account_type, account_name, email):
    """Add an account of type [ACCOUNT-TYPE] with the name [ACCOUNT-NAME]

    \b
    Example:
        $ rucio account add
    """
    args = Arguments({"account_type": account_type, "account": account_name, "email": email})
    add_account(args, ctx.obj.client, ctx.obj.logger, ctx.obj.console, ctx.obj.spinner)


@account.command("list")
@click.option("--type", "type_", type=click.Choice(["USER", "GROUP", "SERVICE"]))
@click.option("--id", help="Filter by identity (e.g. DN)")
@click.option("--filter", help="Filter arguments in form `key=value,another_key=next_value`")  # TODO Explicit numeration of these possible keys
@click.pass_context
def list_(ctx, type_, id, filter):
    """List all accounts that match given filters"""
    args = Arguments({"account_type": type_, "identity": id, "filters": filter})
    list_accounts(args, ctx.obj.client, ctx.obj.logger, ctx.obj.console, ctx.obj.spinner)


@account.command("show")
@click.argument("account-name")
@click.pass_context
def show(ctx, account_name):
    """
    Show info about a single account
    """
    info_account(Arguments({"account": account_name}), ctx.obj.client, ctx.obj.logger, ctx.obj.console, ctx.obj.spinner)


@account.command("remove")
@click.argument("account-name")
@click.pass_context
def remove(ctx, account_name):
    """
    Remove an account
    (WARNING: Permanently disables the account. If you want to temporarily disable, use `account update [account-name] --ban`)
    """
    delete_account(Arguments({"account": account_name}), ctx.obj.client, ctx.obj.logger, ctx.obj.console, ctx.obj.spinner)


@account.command("update")
@click.argument("account-name")
@click.option("--email", help="Email address associated with the account")
@click.option("--ban/--unban", default=None, help="Temporarily disable/enable an account")
@click.pass_context
def update(ctx, ban, account_name, email):
    """Update account settings"""
    args = Arguments({"account": account_name, "key": "email", "value": email})
    if ban is not None:
        if ban:
            ban_account(args, ctx.obj.client, ctx.obj.logger, ctx.obj.console, ctx.obj.spinner)
        else:
            unban_account(args, ctx.obj.client, ctx.obj.logger, ctx.obj.console, ctx.obj.spinner)
    else:
        update_account(args, ctx.obj.client, ctx.obj.logger, ctx.obj.console, ctx.obj.spinner)


@account.group()
def attribute():
    """View or modify account attributes"""


@attribute.command("list")
@click.argument("account-name")
@click.pass_context
def attribute_list(ctx, account_name):
    "List the attributes for a given account"
    list_account_attributes(Arguments({"account": account_name}), ctx.obj.client, ctx.obj.logger, ctx.obj.console, ctx.obj.spinner)


@attribute.command("add")
@click.argument("account-name")
@click.option('--key', help='Attribute key', required=True)
@click.option('--value', help='Attribute value', required=True)
@click.pass_context
def attribute_add(ctx, account_name, key, value):
    """Add a new attribute [key] to an account"""
    add_account_attribute(Arguments({"account": account_name, "key": key, "value": value}), ctx.obj.client, ctx.obj.logger, ctx.obj.console, ctx.obj.spinner)


@attribute.command("remove")
@click.argument("account-name")
@click.option("--key", help="Attribute key", required=True)
@click.pass_context
def attribute_remove(ctx, account_name, key):
    """Remove an attribute from an account without reassigning it"""
    delete_account_attribute(Arguments({"account": account_name, "key": key}), ctx.obj.client, ctx.obj.logger, ctx.obj.console, ctx.obj.spinner)


@account.group("limit")
def limit():
    """View or modify account limits - limit how much data an account can store on an RSE"""


@limit.command("list", help="Shows the space used, the quota limit and the quota left for an account for every RSE where the user have quota.")
@click.argument("account-name")
@click.option("--rse", "--rse-name", help="Show usage for only for this RSE.")
@click.pass_context
def limit_list(ctx, account_name, rse):
    """List the limits and current usage for an account"""
    args = Arguments({"usage_account": account_name, "rse": rse})
    list_account_usage(args, ctx.obj.client, ctx.obj.logger, ctx.obj.console, ctx.obj.spinner)


@limit.command("add")
@click.argument(
    "account-name",
)
@click.option("--rse", "--rse-name", help="Full RSE name", required=True)  # TODO Separate RSE (local) and RSE Expression (global)
@click.option("--bytes", "bytes_", help='Value of the limit; can be specified in bytes ("10000"), with a storage unit ("10GB"), or "infinity"', required=True)
@click.option("--locality", type=click.Choice(["local", "global"]), help="Global or local limit scope", default="local")
@click.pass_context
def limit_add(ctx, account_name, rse, bytes_, locality):
    """Add a new limit for an account on an RSE. An account can have both local and global limits on the same RSE."""
    args = Arguments({"account": account_name, "rse": rse, "bytes": bytes_, "locality": locality})
    set_limits(args, ctx.obj.client, ctx.obj.logger, ctx.obj.console, ctx.obj.spinner)


@limit.command("remove")
@click.argument(
    "account-name",
)
@click.option("--rse", "--rse-name", help="Full RSE name", required=True)
@click.option("--locality", type=click.Choice(["local", "global"]), help="Global or local limit scope", default="local")
@click.pass_context
def limit_remove(ctx, account_name, rse, locality):
    """Remove existing limits for an account on an RSE"""
    args = Arguments({"account": account_name, "rse": rse, "locality": locality})
    delete_limits(args, ctx.obj.client, ctx.obj.logger, ctx.obj.console, ctx.obj.spinner)


@account.group("identity")
def identity():
    """Manage identities for an account - used to login"""


@identity.command("list")
@click.argument("account-name", required=True)
@click.pass_context
def identity_list(ctx, account_name):
    """See all the IDs for [account-name]"""
    args = Arguments({"account": account_name})
    list_identities(args, ctx.obj.client, ctx.obj.logger, ctx.obj.console, ctx.obj.spinner)


@identity.command("add")
@click.argument("account-name", required=True)
@click.option("--type", "type_", type=click.Choice(["X509", "GSS", "USERPASS", "SSH", "SAML", "OIDC"]), help="Authentication type", required=True)
@click.option("--id", help="Identity", required=True)
@click.option("--email", help="Email address associated with the identity", required=True)
@click.option("--password", help="Password if authtype is USERPASS")
@click.pass_context
def identity_add(ctx, account_name, type_, id, email, password):
    """Add a new identity for [account-name]"""
    args = Arguments({"account": account_name, "authtype": type_, "identity": id, "email": email, "password": password})
    legacy_identity_add(args, ctx.obj.client, ctx.obj.logger, ctx.obj.console, ctx.obj.spinner)


@identity.command("remove")
@click.argument("account-name", required=True)
@click.option("--type", "type_", type=click.Choice(["X509", "GSS", "USERPASS", "SSH", "SAML", "OIDC"]), help="Authentication type", required=True)
@click.option("--id", help="Identity", required=True)
@click.pass_context
def identity_remove(ctx, account_name, type_, id):
    """Revoke a given ID's access from an account"""
    args = Arguments({"account": account_name, "authtype": type_, "id": id})
    identity_delete(args, ctx.obj.client, ctx.obj.logger, ctx.obj.console, ctx.obj.spinner)



# Copyright European Organization for Nuclear Research (CERN) since 2012
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

''' add tokens index '''

from alembic import context
from alembic.op import create_foreign_key, create_index, drop_constraint, drop_index

# Alembic revision identifiers
revision = '49a21b4d4357'
down_revision = '2eef46be23d4'


def upgrade():
    '''
    Upgrade the database to this revision
    '''

    if context.get_context().dialect.name in ['oracle', 'mysql', 'postgresql']:
        drop_constraint('TOKENS_ACCOUNT_FK', 'tokens', type_='foreignkey')
        create_index('TOKENS_ACCOUNT_EXPIRED_AT_IDX', 'tokens', ['account', 'expired_at'])
        create_foreign_key('TOKENS_ACCOUNT_FK', 'tokens', 'accounts', ['account'], ['account'])


def downgrade():
    '''
    Downgrade the database to the previous revision
    '''

    if context.get_context().dialect.name in ['oracle', 'mysql', 'postgresql']:
        drop_constraint('TOKENS_ACCOUNT_FK', 'tokens', type_='foreignkey')
        drop_index('TOKENS_ACCOUNT_EXPIRED_AT_IDX', 'tokens')
        create_foreign_key('TOKENS_ACCOUNT_FK', 'tokens', 'accounts', ['account'], ['account'])



# Copyright European Organization for Nuclear Research (CERN) since 2012
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from json import dumps
from typing import TYPE_CHECKING

from flask import Flask, Response, jsonify, request

from rucio.common.exception import (
    AccessDenied,
    CounterNotFound,
    Duplicate,
    InputValidationError,
    InvalidObject,
    InvalidPath,
    InvalidRSEExpression,
    ReplicaNotFound,
    RSEAttributeNotFound,
    RSENotFound,
    RSEOperationNotSupported,
    RSEProtocolDomainNotSupported,
    RSEProtocolNotSupported,
    RSEProtocolPriorityError,
)
from rucio.common.utils import APIEncoder, Availability, render_json
from rucio.gateway.account_limit import get_rse_account_usage
from rucio.gateway.rse import (
    add_distance,
    add_protocol,
    add_qos_policy,
    add_rse,
    add_rse_attribute,
    del_protocols,
    del_rse,
    del_rse_attribute,
    delete_distance,
    delete_qos_policy,
    delete_rse_limits,
    get_distance,
    get_rse,
    get_rse_limits,
    get_rse_protocols,
    get_rse_usage,
    list_qos_policies,
    list_rse_attributes,
    list_rse_usage_history,
    list_rses,
    parse_rse_expression,
    set_rse_limits,
    set_rse_usage,
    update_distance,
    update_protocols,
    update_rse,
)
from rucio.rse import rsemanager
from rucio.web.rest.flaskapi.authenticated_bp import AuthenticatedBlueprint
from rucio.web.rest.flaskapi.v1.common import ErrorHandlingMethodView, check_accept_header_wrapper_flask, generate_http_error_flask, json_parameters, param_get, response_headers, try_stream

if TYPE_CHECKING:
    from rucio.common.types import LFNDict


class RSEs(ErrorHandlingMethodView):
    """ List all RSEs in the database. """

    @check_accept_header_wrapper_flask(['application/x-json-stream'])
    def get(self):
        """
        ---
        summary: List RSEs
        description: Lists all RSEs.
        tags:
          - Rucio Storage Elements
        parameters:
        - name: expression
          in: query
          description: RSE expression to select RSEs.
          schema:
            type: string
        responses:
          200:
            description: OK
            content:
              application/json:
                schema:
                  description: A list with the corresponding rses.
                  type: array
                  items:
                    type: object
                    properties:
                      id:
                        description: The rse id.
                        type: string
                      rse:
                        description: The name of the rse.
                        type: string
                      rse_type:
                        description: The type of the rse.
                        type: string
                      deterministic:
                        description: If the rse is deterministic.
                        type: boolean
                      volatile:
                        description: If the rse is volatile.
                        type: boolean
                      staging_area:
                        description: Is this rse a staging area?
                        type: boolean
                      city:
                        description: The city of the rse.
                        type: string
                      region_code:
                        description: The region_code of the rse.
                        type: string
                      country_name:
                        description: The country name of the rse.
                        type: string
                      continent:
                        description: The continent of the rse.
                        type: string
                      time_zone:
                        description: The time zone of the rse.
                        type: string
                      ISP:
                        description: The isp of the rse.
                        type: string
                      ASN:
                        description: The asn of the rse.
                        type: string
                      longitude:
                        description: The longitude of the rse.
                        type: number
                      latitude:
                        description: The latitude of the rse.
                        type: number
                      availability:
                        description: The availability of the rse.
                        type: integer
                        deprecated: true
                      availability_read:
                        description: If the RSE is readable.
                        type: integer
                      availability_write:
                        description: If the RSE is writable.
                        type: integer
                      availability_delete:
                        description: If the RSE is deletable.
                        type: integer
                      usage:
                        description: The usage of the rse.
                        type: integer
                      qos_class:
                        description: The quality of service class.
                        type: string
          400:
            description: Invalid RSE expression
          401:
            description: Invalid Auth Token
          406:
            description: Not acceptable
        """
        expression = request.args.get('expression', default=None)

        if expression:
            try:
                def generate(vo):
                    for rse in parse_rse_expression(expression, vo=vo):
                        yield render_json(rse=rse) + '\n'

                return try_stream(generate(vo=request.environ.get('vo')))
            except (InvalidRSEExpression, InvalidObject) as error:
                return generate_http_error_flask(400, error)
        else:
            def generate(vo):
                for rse in list_rses(vo=vo):
                    rse['availability'] = Availability(rse['availability_read'], rse['availability_write'], rse['availability_delete']).integer
                    yield render_json(**rse) + '\n'

            return try_stream(generate(vo=request.environ.get('vo')))


class RSE(ErrorHandlingMethodView):
    """ Create, update, get and disable RSE. """

    def post(self, rse):
        """
        ---
        summary: Create RSE
        description: Creates a RSE with all the metadata.
        tags:
          - Rucio Storage Elements
        parameters:
        - name: rse
          in: path
          description: The name of the Rucio Storage Element name.
          schema:
            type: string
          style: simple
        requestBody:
          content:
            application/json:
              schema:
                type: object
                properties:
                  deterministic:
                    description: If the pfn is generated deterministicly.
                    type: boolean
                  volatile:
                    description: RSE cache.
                    type: boolean
                  city:
                    description: The city of the RSE.
                    type: string
                  staging_area:
                    description: Staging area.
                    type: string
                  region_code:
                    description: The region code of the RSE.
                    type: string
                  country_name:
                    description: The country name of the RSE.
                    type: string
                  continent:
                    description: The continent of the RSE.
                    type: string
                  time_zone:
                    description: The time zone of the RSE.
                    type: string
                  ISP:
                    description: The internet service provider of the RSE.
                    type: string
                  rse_type:
                    description: The rse type.
                    type: string
                    enum: ["DISK", "TAPE"]
                  latitude:
                    description: The latitude of the RSE.
                    type: number
                  longitude:
                    description: The longitude of the RSE.
                    type: number
                  ASN:
                    description: The access service network of the RSE.
                    type: string
                  availability:
                    description: The availability of the RSE.
                    type: integer
                    deprecated: true
                  availability_read:
                    description: If the RSE is readable.
                    type: boolean
                  availability_write:
                    description: If the RSE is writable.
                    type: boolean
                  availability_delete:
                    description: If the RSE is deletable.
                    type: boolean
        responses:
          201:
            description: OK
            content:
              application/json:
                schema:
                  type: string
                  enum: ["Created"]
          400:
            description: Cannot decode json parameter dictionary
          401:
            description: Invalid Auth Token
          404:
            description: RSE not found
          409:
            description: RSE already exists.
        """
        kwargs = {
            'deterministic': True,
            'volatile': False,
            'city': None,
            'staging_area': False,
            'region_code': None,
            'country_name': None,
            'continent': None,
            'time_zone': None,
            'ISP': None,
            'rse_type': None,
            'latitude': None,
            'longitude': None,
            'ASN': None,
            'availability_read': None,
            'availability_write': None,
            'availability_delete': None,
        }
        if request.get_data(as_text=True):
            parameters = json_parameters()

            if "availability" in parameters and ("availability_read" in parameters or "availability_write" in parameters or "availability_delete" in parameters):
                return generate_http_error_flask(422, InputValidationError("'availability' can not be specified with 'availability_read', 'availability_write' or 'availability_delete'"))  # noqa: E501

            if "availability" in parameters:
                availability = Availability.from_integer(parameters["availability"])
                kwargs["availability_read"] = availability.read
                kwargs["availability_write"] = availability.write
                kwargs["availability_delete"] = availability.delete

            for keyword in kwargs.keys():
                kwargs[keyword] = param_get(parameters, keyword, default=kwargs[keyword])

        kwargs['issuer'] = request.environ.get('issuer')
        kwargs['vo'] = request.environ.get('vo')
        try:
            add_rse(rse, **kwargs)
        except InvalidObject as error:
            return generate_http_error_flask(400, error)
        except AccessDenied as error:
            return generate_http_error_flask(401, error)
        except RSENotFound as error:
            return generate_http_error_flask(404, error)
        except Duplicate as error:
            return generate_http_error_flask(409, error)

        return 'Created', 201

    def put(self, rse):
        """
        ---
        summary: Update RSE
        description: Update RSE properties.
        tags:
          - Rucio Storage Elements
        parameters:
        - name: rse
          in: path
          description: The name of the Rucio Storage Element name.
          schema:
            type: string
          style: simple
        requestBody:
          content:
            application/json:
              schema:
                type: object
                properties:
                  availability_read:
                    description: The vailability of the RSE.
                    type: boolean
                  availability_write:
                    description: The vailability of the RSE.
                    type: boolean
                  availability_delete:
                    description: The vailability of the RSE.
                    type: boolean
                  deterministic:
                    description: If the pfn is generated deterministicly.
                    type: boolean
                  volatile:
                    description: RSE cache.
                    type: boolean
                  city:
                    description: The city of the RSE.
                    type: string
                  staging_area:
                    description: Staging area.
                    type: string
                  region_code:
                    description: The region code of the RSE.
                    type: string
                  country_name:
                    description: The country name of the RSE.
                    type: string
                  time_zone:
                    description: The time zone of the RSE.
                    type: string
                  rse_type:
                    description: The rse type.
                    type: string
                    enum: ["DISK", "TAPE"]
                  latitude:
                    description: The latitude of the RSE.
                    type: number
                  longitude:
                    description: The longitude of the RSE.
                    type: number
        responses:
          201:
            description: OK
            content:
              application/json:
                schema:
                  type: string
                  enum: ["Created"]
          400:
            description: Cannot decode json parameter dictionary or invalid option provided
          401:
            description: Invalid Auth Token
          404:
            description: RSE not found
        """
        kwargs = {
            'parameters': json_parameters(optional=True),
            'issuer': request.environ.get('issuer'),
            'vo': request.environ.get('vo'),
        }
        try:
            update_rse(rse, **kwargs)
        except (InvalidObject, InputValidationError) as error:
            return generate_http_error_flask(400, error)
        except AccessDenied as error:
            return generate_http_error_flask(401, error)
        except RSENotFound as error:
            return generate_http_error_flask(404, error)
        except Duplicate as error:
            return generate_http_error_flask(409, error)

        return 'Created', 201

    @check_accept_header_wrapper_flask(['application/json'])
    def get(self, rse):
        """
        ---
        summary: Get RSE
        description: Get details about a specific RSE.
        tags:
          - Rucio Storage Elements
        parameters:
        - name: rse
          in: path
          description: The name of the Rucio Storage Element name.
          schema:
            type: string
          style: simple
        responses:
          200:
            description: OK
            content:
              application/json:
                schema:
                  description: The RSE properties.
                  type: object
                  properties:
                    deterministic:
                      description: If the pfn is generated deterministicly.
                      type: boolean
                    volatile:
                      description: RSE cache.
                      type: boolean
                    city:
                      description: The city of the RSE.
                      type: string
                    staging_area:
                      description: Staging area.
                      type: string
                    region_code:
                      description: The region code of the RSE.
                      type: string
                    country_name:
                      description: The country name of the RSE.
                      type: string
                    continent:
                      description: The continent of the RSE.
                      type: string
                    time_zone:
                      description: The time zone of the RSE.
                      type: string
                    ISP:
                      description: The internet service provider of the RSE.
                      type: string
                    rse_type:
                      description: The rse type.
                      type: string
                      enum: ["DISK", "TAPE"]
                    latitude:
                      description: The latitude of the RSE.
                      type: number
                    longitude:
                      description: The longitude of the RSE.
                      type: number
                    ASN:
                      description: The access service network of the RSE.
                      type: string
                    availability:
                      description: The availability of the RSE.
                      type: integer
                      deprecated: true
                    availability_read:
                      description: If the RSE is readable.
                      type: integer
                    availability_write:
                      description: If the RSE is writable.
                      type: integer
                    availability_delete:
                      description: If the RSE is deletable.
          401:
            description: Invalid Auth Token
          404:
            description: RSE not found
          406:
            description: Not acceptable
        """
        try:
            rse = get_rse(rse=rse, vo=request.environ.get('vo'))
            rse['availability'] = Availability(rse['availability_read'], rse['availability_write'], rse['availability_delete']).integer
            return Response(render_json(**rse), content_type="application/json")
        except RSENotFound as error:
            return generate_http_error_flask(404, error)

    def delete(self, rse):
        """
        ---
        summary: Disable RSE
        description: Disable a specific RSE.
        tags:
          - Rucio Storage Elements
        parameters:
        - name: rse
          in: path
          description: The name of the Rucio Storage Element name.
          schema:
            type: string
          style: simple
        responses:
          200:
            description: OK
          401:
            description: Invalid Auth Token
          404:
            description: RSE not found
        """
        try:
            del_rse(rse=rse, issuer=request.environ.get('issuer'), vo=request.environ.get('vo'))
        except (RSENotFound, RSEOperationNotSupported, CounterNotFound) as error:
            return generate_http_error_flask(404, error)
        except AccessDenied as error:
            return generate_http_error_flask(401, error)

        return '', 200


class Attributes(ErrorHandlingMethodView):
    """ Create, update, get and disable RSE attribute."""

    def post(self, rse, key):
        """
        ---
        summary: Create RSE Attribute
        description: Create a RSE attribute with given RSE name.
        tags:
          - Rucio Storage Elements
        parameters:
        - name: rse
          in: path
          description: The name of the Rucio Storage Element name.
          schema:
            type: string
          style: simple
        - name: key
          in: path
          description: The name of the attribute of the RSE.
          schema:
            type: string
          style: simple
        requestBody:
          content:
            application/json:
              schema:
                type: object
                required:
                - value
                properties:
                  value:
                    description: The value of the RSE attribute.
                    type: string
        responses:
          201:
            description: OK
            content:
              application/json:
                schema:
                  type: string
                  enum: ["Created"]
          400:
            description: Cannot decode json parameter dictionary
          401:
            description: Invalid Auth Token
          404:
            description: RSE not found
          409:
            description: Attribute already exists
        """
        parameters = json_parameters()
        value = param_get(parameters, 'value')
        try:
            add_rse_attribute(rse=rse, key=key, value=value, issuer=request.environ.get('issuer'), vo=request.environ.get('vo'))
        except AccessDenied as error:
            return generate_http_error_flask(401, error)
        except Duplicate as error:
            return generate_http_error_flask(409, error)
        except RSENotFound as error:
            return generate_http_error_flask(404, error)

        return 'Created', 201

    @check_accept_header_wrapper_flask(['application/json'])
    def get(self, rse):
        """
        ---
        summary: Get RSE Attributes
        description: Lists all RSE attributes for a RSE.
        tags:
          - Rucio Storage Elements
        parameters:
        - name: rse
          in: path
          description: The name of the Rucio Storage Element name.
          schema:
            type: string
          style: simple
        responses:
          200:
            description: OK
            content:
              application/json:
                schema:
                  description: The RSE attribute list. Returns a dictionary with the attribute names as keys and the values as values.
                  type: object
                  additionalProperties:
                    type: string
          401:
            description: Invalid Auth Token
          404:
            description: RSE not found
          406:
            description: Not acceptable
        """
        try:
            rse_attr = list_rse_attributes(rse, vo=request.environ.get('vo'))
        except AccessDenied as error:
            return generate_http_error_flask(401, error)
        except RSENotFound as error:
            return generate_http_error_flask(404, error)

        return jsonify(rse_attr)

    def delete(self, rse, key):
        """
        ---
        summary: Delete RSE Attribute
        description: Delete an RSE attribute for given RSE name.
        tags:
          - Rucio Storage Elements
        parameters:
        - name: rse
          in: path
          description: The name of the Rucio Storage Element name.
          schema:
            type: string
          style: simple
        - name: key
          in: path
          description: The name of the attribute of the RSE.
          schema:
            type: string
          style: simple
        responses:
          200:
            description: OK
          401:
            description: Invalid Auth Token
          404:
            description: RSE or RSE attribute not found
        """
        try:
            del_rse_attribute(rse=rse, key=key, issuer=request.environ.get('issuer'), vo=request.environ.get('vo'))
        except AccessDenied as error:
            return generate_http_error_flask(401, error)
        except (RSENotFound, RSEAttributeNotFound) as error:
            return generate_http_error_flask(404, error)

        return '', 200


class ProtocolList(ErrorHandlingMethodView):
    """ List supported protocols. """

    @check_accept_header_wrapper_flask(['application/json'])
    def get(self, rse):
        """
        ---
        summary: List RSE Protocols
        description: List all supported protocols of the given RSE.
        tags:
          - Rucio Storage Elements
        parameters:
        - name: rse
          in: path
          description: The name of the Rucio Storage Element name.
          schema:
            type: string
          style: simple
        responses:
          200:
            description: OK
            content:
              application/json:
                schema:
                  description: Supported RSE Protocols and other information.
                  type: object
                  properties:
                    deterministic:
                      description: If the pfn is generated deterministicly.
                      type: boolean
                    volatile:
                      description: RSE cache.
                      type: boolean
                    staging_area:
                      description: Staging area.
                      type: string
                    rse_type:
                      description: The rse type.
                      type: string
                      enum: ["DISK", "TAPE"]
                    availability_read:
                      description: The read availability of the RSE.
                      type: boolean
                    availability_write:
                      description: The write availability of the RSE.
                      type: boolean
                    availability_delete:
                      description: The delete availability of the RSE.
                      type: boolean
                    credentials:
                      description: The credentials, currently None.
                      type: string
                    domain:
                      description: The domains of the RSE protocols.
                      type: array
                    id:
                      description: The RSE id.
                      type: string
                    lfn2pfn_algorithm:
                      description: The algorithm used to translate the logical file names to the physical ones.
                      type: string
                    qos_class:
                      description: The qos class of the RSE.
                      type: string
                    rse:
                      description: The name of the RSE.
                      type: string
                    sign_url:
                      description: The sign url of the RSE.
                      type: string
                    verify_checksum:
                      description: If the checksum of the files should be verified.
                      type: boolean
                    protocols:
                      description: All supported protocols of the RSE.
                      type: array
                      items:
                        type: object
                        description: A supported RSE protocol.
                        properties:
                          hostname:
                            description: The hostname of the protocol.
                            type: string
                          scheme:
                            description: The scheme of the protocol.
                            type: string
                          port:
                            description: The port of the protocol.
                            type: integer
                          prefix:
                            description: The prefix of the protocol.
                            type: string
                          impl:
                            description: The implementation of the protocol.
                            type: string
                          domains:
                            description: The domains of the protocol.
                            type: object
                            properties:
                              lan:
                                description: The lan domain
                                type: object
                                properties:
                                  read:
                                    description: The read value of the lan protocol.
                                    type: integer
                                  write:
                                    description: The write value of the lan protocol.
                                    type: integer
                                  delete:
                                    description: The delete value of the lan protocol.
                                    type: integer
                              wan:
                                description: The wan domain
                                type: object
                                properties:
                                  read:
                                    description: The read value of the wan protocol.
                                    type: integer
                                  write:
                                    description: The read value of the wan protocol.
                                    type: integer
                                  delete:
                                    description: The read value of the wan protocol.
                                    type: integer
                                  third_party_copy_read:
                                    description: The third party copy read value of the wan protocol.
                                    type: integer
                                  third_party_copy_write:
                                    description: The third party copy write value of the wan protocol.
                                    type: integer
                          extended_attributes:
                            description: The extended attributes of the protocol.
                            type: string
          401:
            description: Invalid Auth Token
          404:
            description: RSE not found or RSE Operation, RSE Protocol Domain, RSE Protocol not supported
          406:
            description: Not acceptable
        """
        try:
            p_list = get_rse_protocols(rse, issuer=request.environ.get('issuer'), vo=request.environ.get('vo'))
        except (RSEOperationNotSupported, RSENotFound, RSEProtocolNotSupported, RSEProtocolDomainNotSupported) as error:
            return generate_http_error_flask(404, error)

        if len(p_list['protocols']):
            return jsonify(p_list['protocols'])
        else:
            return generate_http_error_flask(404, RSEProtocolNotSupported.__name__, 'No protocols found for this RSE')


class LFNS2PFNS(ErrorHandlingMethodView):
    """ Translate one-or-more LFNs to corresponding PFNs. """

    @check_accept_header_wrapper_flask(['application/json'])
    def get(self, rse):
        """
        ---
        summary: Translate LFNs to PFNs
        description: Return PFNs for a set of LFNs.  Formatted as a JSON object where the key is a LFN and the value is the corresponding PFN.
        tags:
          - Rucio Storage Elements
        parameters:
        - name: rse
          in: path
          description: The name of the Rucio Storage Element name.
          schema:
            type: string
          style: simple
        - name: lfn
          in: query
          description: The lfns of the request.
          schema:
            type: string
          required: True
        - name: scheme
          in: query
          description: Optional argument to help with the protocol selection (e.g., http / gsiftp / srm)
          schema:
            type: string
        - name: domain
          in: query
          description: Optional argument used to select the protocol for wan or lan use cases.
          schema:
            type: string
        - name: operation
          in: query
          description: Optional query argument to select the protocol for read-vs-writes.
          schema:
            type: string
        responses:
          200:
            description: OK
            content:
              application/json:
                schema:
                  description: The PFNs to the LFNs. Dictionary with lfns as keys and pfns as values.
                  type: object
                  additionalProperties:
                    type:
                      string
          401:
            description: Invalid Auth Token
          404:
            description: RSE not found or RSE Protocol or RSE Protocol Domain not supported
          406:
            description: Not acceptable
        """
        lfns = request.args.getlist('lfn')
        lfns = list(map(lambda lfn: lfn.split(":", 1), lfns))
        if any(filter(lambda info: len(info) != 2, lfns)):
            invalid_lfns = ', '.join(filter(lambda info: len(info) != 2, lfns))
            return generate_http_error_flask(400, InvalidPath.__name__, 'LFN(s) in invalid format: ' + invalid_lfns)
        lfns_list: list["LFNDict"] = list(map(lambda info: {'scope': info[0], 'name': info[1]}, lfns))
        scheme = request.args.get('scheme', default=None)
        domain = request.args.get('domain', default='wan')
        operation = request.args.get('operation', default='write')

        try:
            rse_settings = get_rse_protocols(rse, issuer=request.environ.get('issuer'), vo=request.environ.get('vo'))
        except (RSENotFound, RSEProtocolNotSupported, RSEProtocolDomainNotSupported) as error:
            return generate_http_error_flask(404, error)

        pfns = rsemanager.lfns2pfns(rse_settings, lfns_list, operation=operation, scheme=scheme, domain=domain)
        if not pfns:
            return generate_http_error_flask(404, ReplicaNotFound.__name__, 'No replicas found')

        return jsonify(pfns)


class Protocol(ErrorHandlingMethodView):
    """ Create, Update, Read and delete a specific protocol. """

    def post(self, rse, scheme):
        """
        ---
        summary: Create RSE Protocol
        description: Create a protocol for a given RSE.
        tags:
          - Rucio Storage Elements
        parameters:
        - name: rse
          in: path
          description: The name of the Rucio Storage Element name.
          schema:
            type: string
          style: simple
        - name: scheme
          in: path
          description: The protocol identifier.
          schema:
            type: string
          style: simple
          required: False
        requestBody:
          content:
            application/json:
              schema:
                type: object
                properties:
                  domains:
                    description: The domains for the protocol.
                    type: array
                  port:
                    description: The port the protocol uses.
                    type: integer
                  hostname:
                    description: The hostname of the protocol.
                    type: string
                  extended_attributes:
                    description: Extended attributes for the protocol.
                    type: string
                  prefix:
                    description: The prefix of the Protocol.
                    type: string
                  impl:
                    description: The impl used by the Protocol.
                    type: string
                  read_lan:
                    description: If the protocol is readable via lan.
                    type: integer
                  write_lan:
                    description: If the protocol is writable via lan.
                    type: integer
                  delete_lan:
                    description: If the protocol is deletable via lan.
                    type: integer
                  read_wan:
                    description: If the protocol is readable via wan.
                    type: integer
                  write_wan:
                    description: If the protocol is writable via wan.
                    type: integer
                  delete_wan:
                    description: If the protocol is deletable via wan.
                    type: integer
        responses:
          201:
            description: OK
            content:
              application/json:
                schema:
                  type: string
                  enum: ["Created"]
          400:
            description: Cannot decode json parameter dictionary
          401:
            description: Invalid Auth Token
          404:
            description: RSE not found or RSE Protocol Domain not supported
          409:
            description: RSE protocol priority error
        """
        parameters = json_parameters()

        # Fill defaults and check mandatory parameters
        parameters['scheme'] = scheme

        try:
            add_protocol(rse, issuer=request.environ.get('issuer'), vo=request.environ.get('vo'), data=parameters)
        except (RSENotFound, RSEProtocolDomainNotSupported) as error:
            return generate_http_error_flask(404, error)
        except AccessDenied as error:
            return generate_http_error_flask(401, error)
        except (Duplicate, RSEProtocolPriorityError) as error:
            return generate_http_error_flask(409, error)
        except InvalidObject as error:
            return generate_http_error_flask(400, error)

        return 'Created', 201

    @check_accept_header_wrapper_flask(['application/json'])
    def get(self, rse, scheme):
        """
        ---
        summary: Get Protocols
        description: List all references of the provided RSE for the given protocol.
        tags:
          - Rucio Storage Elements
        parameters:
        - name: rse
          in: path
          description: The name of the Rucio Storage Element name.
          schema:
            type: string
          style: simple
        - name: scheme
          in: path
          description: The protocol identifier.
          schema:
            type: string
          style: simple
          required: False
        responses:
          200:
            description: OK
            content:
              application/json:
                schema:
                  description: A dict with RSE information and supported protocols.
                  type: object
                  properties:
                    deterministic:
                      description: If the pfn is generated deterministicly.
                      type: boolean
                    volatile:
                      description: RSE cache.
                      type: boolean
                    staging_area:
                      description: Staging area.
                      type: string
                    rse_type:
                      description: The rse type.
                      type: string
                      enum: ["DISK", "TAPE"]
                    availability_read:
                      description: The read availability of the RSE.
                      type: boolean
                    availability_write:
                      description: The write availability of the RSE.
                      type: boolean
                    availability_delete:
                      description: The delete availability of the RSE.
                      type: boolean
                    credentials:
                      description: The credentials, currently None.
                      type: string
                    domain:
                      description: The domains of the RSE protocols.
                      type: array
                    id:
                      description: The RSE id.
                      type: string
                    lfn2pfn_algorithm:
                      description: The algorithm used to translate the logical file names to the physical ones.
                      type: string
                    qos_class:
                      description: The qos class of the RSE.
                      type: string
                    rse:
                      description: The name of the RSE.
                      type: string
                    sign_url:
                      description: The sign url of the RSE.
                      type: string
                    verify_checksum:
                      description: If the checksum of the files should be verified.
                      type: boolean
                    protocols:
                      description: All supported protocols of the RSE.
                      type: array
                      items:
                        type: object
                        description: A supported RSE protocol.
                        properties:
                          hostname:
                            description: The hostname of the protocol.
                            type: string
                          scheme:
                            description: The scheme of the protocol.
                            type: string
                          port:
                            description: The port of the protocol.
                            type: integer
                          prefix:
                            description: The prefix of the protocol.
                            type: string
                          impl:
                            description: The implementation of the protocol.
                            type: string
                          domains:
                            description: The domains of the protocol.
                            type: object
                            properties:
                              lan:
                                description: The lan domain
                                type: object
                                properties:
                                  read:
                                    description: The read value of the lan protocol.
                                    type: integer
                                  write:
                                    description: The write value of the lan protocol.
                                    type: integer
                                  delete:
                                    description: The delete value of the lan protocol.
                                    type: integer
                              wan:
                                description: The wan domain
                                type: object
                                properties:
                                  read:
                                    description: The read value of the wan protocol.
                                    type: integer
                                  write:
                                    description: The read value of the wan protocol.
                                    type: integer
                                  delete:
                                    description: The read value of the wan protocol.
                                    type: integer
                                  third_party_copy_read:
                                    description: The third party copy read value of the wan protocol.
                                    type: integer
                                  third_party_copy_write:
                                    description: The third party copy write value of the wan protocol.
                                    type: integer
                          extended_attributes:
                            description: The extended attributes of the protocol.
                            type: string
          401:
            description: Invalid Auth Token
          404:
            description: RSE not found or Protocol or Protocol domain not Supported.
          406:
            description: Not acceptable
        """
        try:
            p_list = get_rse_protocols(rse, issuer=request.environ.get('issuer'), vo=request.environ.get('vo'))
        except (RSENotFound, RSEProtocolNotSupported, RSEProtocolDomainNotSupported) as error:
            return generate_http_error_flask(404, error)

        return jsonify(p_list)

    def put(self, rse, scheme, hostname=None, port=None):
        """
        ---
        summary: Update Protocol Attributes
        description: Updates attributes of an existing protocol entry. Because protocol identifier, hostname, and port are used as unique identifier they are immutable.
        tags:
          - Rucio Storage Elements
        parameters:
        - name: rse
          in: path
          description: The name of the Rucio Storage Element name.
          schema:
            type: string
          style: simple
        - name: scheme
          in: path
          description: The protocol identifier.
          schema:
            type: string
          style: simple
        - name: hostname
          in: path
          description: The hostname of the protocol.
          schema:
            type: string
          style: simple
          required: False
        - name: port
          in: path
          description: The port of the protocol.
          schema:
            type: integer
          style: simple
          required: False
        responses:
          200:
            description: OK
            content:
              application/json:
                schema:
                  description: A dict with RSE information and supported protocols.
                  type: object
                  properties:
                    deterministic:
                      description: If the pfn is generated deterministicly.
                      type: boolean
                    volatile:
                      description: RSE cache.
                      type: boolean
                    staging_area:
                      description: Staging area.
                      type: string
                    rse_type:
                      description: The rse type.
                      type: string
                      enum: ["DISK", "TAPE"]
                    availability_read:
                      description: The read availability of the RSE.
                      type: boolean
                    availability_write:
                      description: The write availability of the RSE.
                      type: boolean
                    availability_delete:
                      description: The delete availability of the RSE.
                      type: boolean
                    credentials:
                      description: The credentials, currently None.
                      type: string
                    domain:
                      description: The domains of the RSE protocols.
                      type: array
                    id:
                      description: The RSE id.
                      type: string
                    lfn2pfn_algorithm:
                      description: The algorithm used to translate the logical file names to the physical ones.
                      type: string
                    qos_class:
                      description: The qos class of the RSE.
                      type: string
                    rse:
                      description: The name of the RSE.
                      type: string
                    sign_url:
                      description: The sign url of the RSE.
                      type: string
                    verify_checksum:
                      description: If the checksum of the files should be verified.
                      type: boolean
                    protocols:
                      description: All supported protocols of the RSE.
                      type: array
                      items:
                        type: object
                        description: A supported RSE protocol.
                        properties:
                          hostname:
                            description: The hostname of the protocol.
                            type: string
                          scheme:
                            description: The scheme of the protocol.
                            type: string
                          port:
                            description: The port of the protocol.
                            type: integer
                          prefix:
                            description: The prefix of the protocol.
                            type: string
                          impl:
                            description: The implementation of the protocol.
                            type: string
                          domains:
                            description: The domains of the protocol.
                            type: object
                            properties:
                              lan:
                                description: The lan domain
                                type: object
                                properties:
                                  read:
                                    description: The read value of the lan protocol.
                                    type: integer
                                  write:
                                    description: The write value of the lan protocol.
                                    type: integer
                                  delete:
                                    description: The delete value of the lan protocol.
                                    type: integer
                              wan:
                                description: The wan domain
                                type: object
                                properties:
                                  read:
                                    description: The read value of the wan protocol.
                                    type: integer
                                  write:
                                    description: The read value of the wan protocol.
                                    type: integer
                                  delete:
                                    description: The read value of the wan protocol.
                                    type: integer
                                  third_party_copy_read:
                                    description: The third party copy read value of the wan protocol.
                                    type: integer
                                  third_party_copy_write:
                                    description: The third party copy write value of the wan protocol.
                                    type: integer
                          extended_attributes:
                            description: The extended attributes of the protocol.
                            type: string
          401:
            description: Invalid Auth Token
          404:
            description: RSE not found or Protocol or Protocol domain not Supported.
          406:
            description: Not acceptable
        """
        parameters = json_parameters()
        try:
            update_protocols(
                rse,
                issuer=request.environ.get('issuer'),
                vo=request.environ.get('vo'),
                scheme=scheme,
                hostname=hostname,
                port=int(port) if port else None,
                data=parameters,
            )
        except InvalidObject as error:
            return generate_http_error_flask(400, error)
        except (RSEProtocolNotSupported, RSENotFound, RSEProtocolDomainNotSupported) as error:
            return generate_http_error_flask(404, error)
        except (RSEProtocolPriorityError, Duplicate) as error:
            return generate_http_error_flask(409, error)

        return '', 200

    def delete(self, rse, scheme, hostname=None, port=None):
        """
        ---
        summary: Delete Protocol Attributes
        description: Delete all protocol attributes.
        tags:
          - Rucio Storage Elements
        parameters:
        - name: rse
          in: path
          description: The name of the Rucio Storage Element name.
          schema:
            type: string
          style: simple
        - name: scheme
          in: path
          description: The protocol identifier.
          schema:
            type: string
          style: simple
        - name: hostname
          in: path
          description: The hostname of the protocol.
          schema:
            type: string
          style: simple
          required: False
        - name: port
          in: path
          description: The port of the protocol.
          schema:
            type: integer
          style: simple
          required: False
        responses:
          200:
            description: OK
          401:
            description: Invalid Auth Token
          404:
            description: Rse not found or protocol not supported
        """
        try:
            del_protocols(
                rse,
                issuer=request.environ.get('issuer'),
                vo=request.environ.get('vo'),
                scheme=scheme,
                hostname=hostname,
                port=int(port) if port else None
            )
        except (RSEProtocolNotSupported, RSENotFound) as error:
            return generate_http_error_flask(404, error)

        return '', 200


class Usage(ErrorHandlingMethodView):
    """ Update and read RSE space usage information. """

    @check_accept_header_wrapper_flask(['application/x-json-stream'])
    def get(self, rse):
        """
        ---
        summary: Get Rse Usage Information
        description: Get rse usage information.
        tags:
          - Rucio Storage Elements
        parameters:
        - name: rse
          in: path
          description: The name of the Rucio Storage Element name.
          schema:
            type: string
          style: simple
        - name: per_account
          in: query
          description: Boolean whether the usage should be also calculated per account or not.
          schema:
            type: boolean
        - name: source
          in: query
          description: The information source, e.g., srm.
          schema:
            type: string
        responses:
          200:
            description: OK
            content:
              application/x-json-stream:
                schema:
                  description: A list with the rse usage.
                  type: array
                  items:
                    type: object
                    properties:
                      rse_id:
                        description: The id of the rse.
                        type: string
                      rse:
                        description: The name of the rse.
                        type: string
                      source:
                        description: The source of the rse.
                        type: string
                      used:
                        description: The number of used bytes.
                        type: integer
                      free:
                        description: The number of free bytes.
                        type: integer
                      total:
                        description: The number of total bytes.
                        type: integer
                      files:
                        description: The number of files.
                        type: integer
                      updated_at:
                        description: The last time it got updated.
                        type: string
          401:
            description: Invalid Auth Token
          404:
            description: Rse not found
          406:
            description: Not acceptable
        """
        per_account = request.args.get('per_account') == 'True'
        try:
            def generate(issuer, source, per_account, vo):
                for usage in get_rse_usage(rse, issuer=issuer, source=source, per_account=per_account, vo=vo):
                    yield render_json(**usage) + '\n'

            return try_stream(
                generate(
                    issuer=request.environ.get('issuer'),
                    source=request.args.get('source'),
                    per_account=per_account,
                    vo=request.environ.get('vo'),
                )
            )
        except RSENotFound as error:
            return generate_http_error_flask(404, error)

    def put(self, rse):
        """
        ---
        summary: Update Rse Usage
        description: Update the RSE Update information.
        tags:
          - Rucio Storage Elements
        parameters:
        - name: rse
          in: path
          description: The name of the Rucio Storage Element name.
          schema:
            type: string
          style: simple
        requestBody:
          content:
            application/json:
              schema:
                type: object
                properties:
                  source:
                    description: The information source, e.g. srm.
                    type: string
                  used:
                    description: The number of used bytes.
                    type: integer
                  free:
                    description: The number of free bytes.
                    type: integer
                  files:
                    description: The number of files.
                    type: integer
        responses:
          200:
            description: OK
          400:
            description: Can not decode json parameter list.
          401:
            description: Invalid Auth Token
          404:
            description: Rse not found
          406:
            description: Not acceptable
        """
        parameters = json_parameters()
        kwargs = {'source': None, 'used': None, 'free': None, 'files': None}
        for keyword in kwargs.keys():
            kwargs[keyword] = param_get(parameters, keyword, default=kwargs[keyword])

        try:
            set_rse_usage(rse=rse, issuer=request.environ.get('issuer'), vo=request.environ.get('vo'), **kwargs)
        except AccessDenied as error:
            return generate_http_error_flask(401, error)
        except RSENotFound as error:
            return generate_http_error_flask(404, error)

        return '', 200


class UsageHistory(ErrorHandlingMethodView):
    """ Read RSE space usage history information. """

    @check_accept_header_wrapper_flask(['application/x-json-stream'])
    def get(self, rse):
        """
        ---
        summary: Get Rse Usage History
        description: Get the rse usage history
        tags:
          - Rucio Storage Elements
        parameters:
        - name: rse
          in: path
          description: The name of the Rucio Storage Element name.
          schema:
            type: string
          style: simple
        responses:
          200:
            description: OK
            content:
              application/x-json-stream:
                schema:
                  description: A list with the rse usage history items.
                  type: array
                  items:
                    type: object
                    properties:
                      rse_id:
                        description: The id of the rse.
                        type: string
                      rse:
                        description: The name of the rse.
                        type: string
                      source:
                        description: The source of the rse.
                        type: string
                      used:
                        description: The number of used bytes.
                        type: integer
                      free:
                        description: The number of free bytes.
                        type: integer
                      total:
                        description: The number of total bytes.
                        type: integer
                      updated_at:
                        description: The last time it got updated.
                        type: string
          401:
            description: Invalid Auth Token
          404:
            description: Rse not found
          406:
            description: Not acceptable
        """
        try:
            def generate(issuer, source, vo):
                for usage in list_rse_usage_history(rse=rse, issuer=issuer, source=source, vo=vo):
                    yield render_json(**usage) + '\n'

            return try_stream(generate(issuer=request.environ.get('issuer'), source=request.args.get('source'), vo=request.environ.get('vo')))
        except RSENotFound as error:
            return generate_http_error_flask(404, error)


class Limits(ErrorHandlingMethodView):
    """ Create, Update, Read and delete RSE limits. """

    @check_accept_header_wrapper_flask(['application/json'])
    def get(self, rse):
        """
        ---
        summary: Get Rse Limits
        description: Get the rse limits.
        tags:
          - Rucio Storage Elements
        parameters:
        - name: rse
          in: path
          description: The name of the Rucio Storage Element name.
          schema:
            type: string
          style: simple
        responses:
          200:
            description: OK
            content:
              application/json:
                schema:
                  description: The limits.
                  type: object
                  additionalProperties:
                    x-additionalPropertiesName: limit name
                    description: An item with the name as key and the value as value.
                    type: integer
          401:
            description: Invalid Auth Token
          404:
            description: Rse not found
          406:
            description: Not acceptable
        """
        try:
            limits = get_rse_limits(rse=rse, issuer=request.environ.get('issuer'), vo=request.environ.get('vo'))
            return Response(render_json(**limits), content_type="application/json")
        except RSENotFound as error:
            return generate_http_error_flask(404, error)

    def put(self, rse):
        """
        ---
        summary: Update Rse Limit
        description: Update an rse limit.
        tags:
          - Rucio Storage Elements
        parameters:
        - name: rse
          in: path
          description: The name of the Rucio Storage Element name.
          schema:
            type: string
          style: simple
        requestBody:
          content:
            application/json:
              schema:
                type: object
                properties:
                  name:
                    description: The name of the limit.
                    type: string
                  value:
                    description: The value of the limit.
                    type: integer
        responses:
          200:
            description: OK
          401:
            description: Invalid Auth Token
          404:
            description: Rse not found
          406:
            description: Not acceptable
        """
        parameters = json_parameters()
        kwargs = {'name': None, 'value': None}
        for keyword in kwargs.keys():
            kwargs[keyword] = param_get(parameters, keyword, default=kwargs[keyword])

        try:
            set_rse_limits(rse=rse, issuer=request.environ.get('issuer'), vo=request.environ.get('vo'), **kwargs)
        except AccessDenied as error:
            return generate_http_error_flask(401, error)
        except RSENotFound as error:
            return generate_http_error_flask(404, error)

        return '', 200

    def delete(self, rse):
        """
        ---
        summary: Delete Rse Limit
        description: Delete an rse limit
        tags:
          - Rucio Storage Elements
        parameters:
        - name: rse
          in: path
          description: The name of the Rucio Storage Element name.
          schema:
            type: string
          style: simple
        requestBody:
          content:
            application/json:
              schema:
                type: object
                required:
                - name
                properties:
                  name:
                    description: The name of the limit.
                    type: string
        responses:
          200:
            description: OK
          401:
            description: Invalid Auth Token
          404:
            description: Rse not found
          406:
            description: Not acceptable
        """
        parameters = json_parameters()
        name = param_get(parameters, 'name')

        try:
            delete_rse_limits(rse=rse, name=name, issuer=request.environ.get('issuer'), vo=request.environ.get('vo'))
        except AccessDenied as error:
            return generate_http_error_flask(401, error)
        except RSENotFound as error:
            return generate_http_error_flask(404, error)

        return '', 200


class RSEAccountUsageLimit(ErrorHandlingMethodView):
    """ Read and delete RSE limits for accounts. """

    @check_accept_header_wrapper_flask(['application/json'])
    def get(self, rse):
        """
        ---
        summary: Get Rse Account Usage and Limit
        description: Returns the usage and limit of an account for a rse.
        tags:
          - Rucio Storage Elements
        parameters:
        - name: rse
          in: path
          description: The name of the Rucio Storage Element name.
          schema:
            type: string
          style: simple
        responses:
          200:
            description: OK
            content:
              application/json:
                schema:
                  description: A list with the rse account limits and usages.
                  type: array
                  items:
                    type: object
                    properties:
                      rse_id:
                        description: The id of the rse.
                        type: string
                      rse:
                        description: The name of the rse.
                        type: string
                      account:
                        description: The account.
                        type: string
                      used_files:
                        description: The number of used files.
                        type: integer
                      used_bytes:
                        description: The number of used bytes.
                        type: integer
                      quota_bytes:
                        description: The number of quota bytes.
                        type: integer
          401:
            description: Invalid Auth Token
          404:
            description: Rse not found
          406:
            description: Not acceptable
        """
        try:
            def generate(vo):
                for usage in get_rse_account_usage(rse=rse, vo=vo):
                    yield render_json(**usage) + '\n'

            return try_stream(generate(vo=request.environ.get('vo')), content_type='application/json')
        except RSENotFound as error:
            return generate_http_error_flask(404, error)


class Distance(ErrorHandlingMethodView):
    """ Create/Update and read distances between RSEs. """

    @check_accept_header_wrapper_flask(['application/json'])
    def get(self, source, destination):
        """
        ---
        summary: Get Rse Distances
        description: Returns the distances between a source and destination rse.
        tags:
          - Rucio Storage Elements
        parameters:
        - name: source
          in: path
          description: The name of the source Rucio Storage Element.
          schema:
            type: string
          style: simple
        - name: destination
          in: path
          description: The name of the destination Rucio Storage Element.
          schema:
            type: string
          style: simple
        responses:
          200:
            description: OK
            content:
              application/json:
                schema:
                  description: The distances between the Rses.
                  type: array
                  items:
                    type: object
                    description: One distance between source and destination.
                    properties:
                      src_rse_id:
                        description: The source rse id.
                        type: string
                      dest_rse_id:
                        description: The destination rse id.
                        type: string
                      distance:
                        description: The distance between RSEs.
                        type: integer
                      ranking:
                        deprecated: true
                        description: Same as distance.
                        type: integer
          401:
            description: Invalid Auth Token
          404:
            description: Rse not found
          406:
            description: Not acceptable
        """
        try:
            distance = get_distance(source=source, destination=destination, issuer=request.environ.get('issuer'), vo=request.environ.get('vo'))
            return Response(dumps(distance, cls=APIEncoder), content_type="application/json")
        except RSENotFound as error:
            return generate_http_error_flask(404, error)

    def post(self, source, destination):
        """
        ---
        summary: Create Rse Distance
        description: Post a rse distance.
        tags:
          - Rucio Storage Elements
        parameters:
        - name: source
          in: path
          description: The name of the source Rucio Storage Element.
          schema:
            type: string
          style: simple
        - name: destination
          in: path
          description: The name of the destination Rucio Storage Element.
          schema:
            type: string
          style: simple
        requestBody:
          content:
            application/json:
              schema:
                type: object
                properties:
                  distance:
                    description: The distance between RSEs.
                    type: integer
                  ranking:
                    deprecated: true
                    description: Same as distance.
                    type: integer
        responses:
          201:
            description: OK
            content:
              application/json:
                schema:
                  type: string
                  enum: ["Created"]
          401:
            description: Invalid Auth Token
          404:
            description: Rse not found
          406:
            description: Not acceptable
        """
        parameters = json_parameters()

        distance = param_get(parameters, 'distance', default=None)
        if distance is None:
            distance = param_get(parameters, 'ranking', default=None)

        try:
            add_distance(
                source=source,
                destination=destination,
                distance=distance,
                issuer=request.environ.get('issuer'),
                vo=request.environ.get('vo'),
            )
        except AccessDenied as error:
            return generate_http_error_flask(401, error)
        except RSENotFound as error:
            return generate_http_error_flask(404, error)
        except Duplicate as error:
            return generate_http_error_flask(409, error)

        return 'Created', 201

    def put(self, source, destination):
        """
        ---
        summary: Update Rse Distance
        description: Update rse distance information.
        tags:
          - Rucio Storage Elements
        parameters:
        - name: source
          in: path
          description: The name of the source Rucio Storage Element.
          schema:
            type: string
          style: simple
        - name: destination
          in: path
          description: The name of the destination Rucio Storage Element.
          schema:
            type: string
          style: simple
        requestBody:
          content:
            application/json:
              schema:
                type: object
                properties:
                  distance:
                    description: The distance between the RSEs.
                    type: integer
                  ranking:
                    deprecated: true
                    description: Same as distance.
                    type: integer
        responses:
          201:
            description: OK
            content:
              application/json:
                schema:
                  type: string
                  enum: ["Created"]
          401:
            description: Invalid Auth Token
          404:
            description: Rse not found
          406:
            description: Not acceptable
        """
        parameters = json_parameters()

        distance = param_get(parameters, 'distance', default=None)
        if distance is None:
            distance = param_get(parameters, 'ranking', default=None)

        try:
            update_distance(
                source=source,
                destination=destination,
                distance=distance,
                issuer=request.environ.get('issuer'),
                vo=request.environ.get('vo'),
            )
        except AccessDenied as error:
            return generate_http_error_flask(401, error)
        except RSENotFound as error:
            return generate_http_error_flask(404, error)

        return '', 200

    def delete(self, source, destination):
        """
        ---
        summary: Delete Rse Distance
        description: Delete distance information between source RSE and destination RSE.
        tags:
          - Rucio Storage Elements
        parameters:
        - name: source
          in: path
          description: The name of the source Rucio Storage Element.
          schema:
            type: string
          style: simple
        - name: destination
          in: path
          description: The name of the destination Rucio Storage Element.
          schema:
            type: string
          style: simple
        responses:
          200:
            description: OK
            content:
              application/json:
                schema:
                  type: string
                  enum: ["Deleted"]
          401:
            description: Invalid Auth Token
          404:
            description: Rse not found
          406:
            description: Not acceptable
        """
        try:
            delete_distance(
                source=source,
                destination=destination,
                issuer=request.environ.get('issuer'),
                vo=request.environ.get('vo')
            )
        except AccessDenied as error:
            return generate_http_error_flask(401, error)
        except RSENotFound as error:
            return generate_http_error_flask(404, error)

        return 'Deleted', 200


class QoSPolicy(ErrorHandlingMethodView):
    """ Add/Delete/List QoS policies on an RSE. """

    @check_accept_header_wrapper_flask(['application/json'])
    def post(self, rse, policy):
        """
        ---
        summary: Add QoS policy
        description: Add a QoS Policy to a RSE.
        tags:
          - Rucio Storage Elements
        parameters:
        - name: rse
          in: path
          description: The name of the Rucio Storage Element name.
          schema:
            type: string
          style: simple
        - name: policy
          in: path
          description: The QoS policy to add to and rse.
          schema:
            type: string
          style: simple
        responses:
          201:
            description: OK
            content:
              application/json:
                schema:
                  type: string
                  enum: ["Created"]
          401:
            description: Invalid Auth Token
          404:
            description: Rse not found
          406:
            description: Not acceptable
        """
        try:
            add_qos_policy(rse=rse, qos_policy=policy, issuer=request.environ.get('issuer'), vo=request.environ.get('vo'))
        except RSENotFound as error:
            return generate_http_error_flask(404, error)

        return 'Created', 201

    @check_accept_header_wrapper_flask(['application/json'])
    def delete(self, rse, policy):
        """
        ---
        summary: Delete QoS Policy
        description: Delete QoS policy from RSE.
        tags:
          - Rucio Storage Elements
        parameters:
        - name: rse
          in: path
          description: The name of the Rucio Storage Element name.
          schema:
            type: string
          style: simple
        - name: policy
          in: path
          description: The QoS policy to add to and rse.
          schema:
            type: string
          style: simple
        responses:
          200:
            description: OK
          401:
            description: Invalid Auth Token
          404:
            description: Rse not found
          406:
            description: Not acceptable
        """
        try:
            delete_qos_policy(rse=rse, qos_policy=policy, issuer=request.environ.get('issuer'), vo=request.environ.get('vo'))
        except RSENotFound as error:
            return generate_http_error_flask(404, error)

        return '', 200

    @check_accept_header_wrapper_flask(['application/json'])
    def get(self, rse):
        """
        ---
        summary: Gett QoS Policies
        description: List all QoS policies for an Rse.
        tags:
          - Rucio Storage Elements
        parameters:
        - name: rse
          in: path
          description: The name of the Rucio Storage Element name.
          schema:
            type: string
          style: simple
        responses:
          200:
            description: OK
            content:
              application/json:
                schema:
                  description: A list with all the QoS policies for an Rse.
                  type: array
                  items:
                    type: object
                    properties:
                      rse_id:
                        description: The rse id.
                        type: string
                      qos_policy:
                        description: The qos policy.
                        type: string
          401:
            description: Invalid Auth Token
          404:
            description: Rse not found
          406:
            description: Not acceptable
        """
        try:
            qos_policies = list_qos_policies(rse=rse, issuer=request.environ.get('issuer'), vo=request.environ.get('vo'))
            return Response(dumps(qos_policies, cls=APIEncoder), content_type='application/json')
        except RSENotFound as error:
            return generate_http_error_flask(404, error)


def blueprint():
    bp = AuthenticatedBlueprint('rses', __name__, url_prefix='/rses')

    attributes_view = Attributes.as_view('attributes')
    bp.add_url_rule('/<rse>/attr/<key>', view_func=attributes_view, methods=['post', 'delete'])
    bp.add_url_rule('/<rse>/attr/', view_func=attributes_view, methods=['get', ])
    distance_view = Distance.as_view('distance')
    bp.add_url_rule('/<source>/distances/<destination>', view_func=distance_view, methods=['get', 'post', 'put', 'delete'])
    protocol_view = Protocol.as_view('protocol')
    bp.add_url_rule('/<rse>/protocols/<scheme>/<hostname>/<port>', view_func=protocol_view, methods=['delete', 'put'])
    bp.add_url_rule('/<rse>/protocols/<scheme>/<hostname>', view_func=protocol_view, methods=['delete', 'put'])
    bp.add_url_rule('/<rse>/protocols/<scheme>', view_func=protocol_view, methods=['get', 'post', 'delete', 'put'])
    protocol_list_view = ProtocolList.as_view('protocol_list')
    bp.add_url_rule('/<rse>/protocols', view_func=protocol_list_view, methods=['get', ])
    lfns2pfns_view = LFNS2PFNS.as_view('lfns2pfns')
    bp.add_url_rule('/<rse>/lfns2pfns', view_func=lfns2pfns_view, methods=['get', ])
    rse_account_usage_limit_view = RSEAccountUsageLimit.as_view('rse_account_usage_limit')
    bp.add_url_rule('/<rse>/accounts/usage', view_func=rse_account_usage_limit_view, methods=['get', ])
    usage_view = Usage.as_view('usage')
    bp.add_url_rule('/<rse>/usage', view_func=usage_view, methods=['get', 'put'])
    usage_history_view = UsageHistory.as_view('usage_history')
    bp.add_url_rule('/<rse>/usage/history', view_func=usage_history_view, methods=['get', ])
    limits_view = Limits.as_view('limits')
    bp.add_url_rule('/<rse>/limits', view_func=limits_view, methods=['get', 'put', 'delete'])
    qos_policy_view = QoSPolicy.as_view('qos_policy')
    bp.add_url_rule('/<rse>/qos_policy', view_func=qos_policy_view, methods=['get', ])
    bp.add_url_rule('/<rse>/qos_policy/<policy>', view_func=qos_policy_view, methods=['post', 'delete'])
    rse_view = RSE.as_view('rse')
    bp.add_url_rule('/<rse>', view_func=rse_view, methods=['get', 'delete', 'put', 'post'])
    rses_view = RSEs.as_view('rses')
    bp.add_url_rule('/', view_func=rses_view, methods=['get', ])

    bp.after_request(response_headers)
    return bp


def make_doc():
    """ Only used for sphinx documentation """
    doc_app = Flask(__name__)
    doc_app.register_blueprint(blueprint())
    return doc_app



# Copyright European Organization for Nuclear Research (CERN) since 2012
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from copy import deepcopy
from typing import TYPE_CHECKING, Any, Optional

import rucio.gateway.permission
from rucio.common.constants import RESERVED_KEYS
from rucio.common.exception import AccessDenied, InvalidObject, RucioException
from rucio.common.schema import validate_schema
from rucio.common.types import InternalAccount, InternalScope
from rucio.common.utils import gateway_update_return_dict
from rucio.core import did, naming_convention
from rucio.core import meta_conventions as meta_convention_core
from rucio.core.rse import get_rse_id
from rucio.db.sqla.constants import DIDType
from rucio.db.sqla.session import read_session, stream_session, transactional_session

if TYPE_CHECKING:
    from collections.abc import Iterable, Iterator, Mapping, Sequence

    from sqlalchemy.orm import Session


@stream_session
def list_dids(
    scope: str,
    filters: 'Mapping[Any, Any]',
    did_type: str = 'collection',
    ignore_case: bool = False,
    limit: Optional[int] = None,
    offset: Optional[int] = None,
    long: bool = False,
    recursive: bool = False,
    vo: str = 'def',
    *,
    session: "Session"
) -> 'Iterator[dict[str, Any]]':
    """
    List dids in a scope.

    :param scope: The scope name.
    :param filters: Filter arguments in form supported by the filter engine.
    :param did_type:  The type of the did: all(container, dataset, file), collection(dataset or container), dataset, container
    :param ignore_case: Ignore case distinctions.
    :param limit: The maximum number of DIDs returned.
    :param offset: Offset number.
    :param long: Long format option to display more information for each DID.
    :param recursive: Recursively list DIDs content.
    :param vo: The VO to act on.
    :param session: The database session in use.
    """
    internal_scope = InternalScope(scope, vo=vo)

    # replace account and scope in filters with internal representation
    for or_group in filters:
        if 'account' in or_group:
            or_group['account'] = InternalAccount(or_group['account'], vo=vo)
        if 'scope' in or_group:
            or_group['account'] = InternalScope(or_group['scope'], vo=vo)

    result = did.list_dids(scope=internal_scope, filters=filters, did_type=did_type, ignore_case=ignore_case,
                           limit=limit, offset=offset, long=long, recursive=recursive, session=session)

    for d in result:
        yield gateway_update_return_dict(d, session=session)


@transactional_session
def add_did(
    scope: str,
    name: str,
    did_type: str,
    issuer: str,
    account: Optional[str] = None,
    statuses: Optional[dict[str, str]] = None,
    meta: Optional[dict[str, str]] = None,
    rules: Optional['Sequence[dict[str, Any]]'] = None,
    lifetime: Optional[str] = None,
    dids: Optional['Sequence[dict[str, Any]]'] = None,
    rse: Optional[str] = None,
    vo: str = 'def',
    *,
    session: "Session"
) -> None:
    """
    Add data did.

    :param scope: The scope name.
    :param name: The data identifier name.
    :param did_type: The data identifier type.
    :param issuer: The issuer account.
    :param account: The account owner. If None, then issuer is selected as owner.
    :param statuses: Dictionary with statuses, e.g.g {'monotonic':True}.
    :meta: Meta-data associated with the data identifier is represented using key/value pairs in a dictionary.
    :rules: Replication rules associated with the data did. A list of dictionaries, e.g., [{'copies': 2, 'rse_expression': 'TIERS1'}, ].
    :param lifetime: DID's lifetime (in seconds).
    :param dids: The content.
    :param rse: The RSE name when registering replicas.
    :param vo: The VO to act on.
    :param session: The database session in use.
    """
    statuses = statuses or {}
    meta = meta or {}
    rules = rules or []
    dids = dids or []
    v_did = {'name': name, 'type': did_type.upper(), 'scope': scope}
    validate_schema(name='did', obj=v_did, vo=vo)
    validate_schema(name='dids', obj=dids, vo=vo)
    validate_schema(name='rse', obj=rse, vo=vo)
    kwargs = {'scope': scope, 'name': name, 'type': did_type, 'issuer': issuer, 'account': account, 'statuses': statuses, 'meta': meta, 'rules': rules, 'lifetime': lifetime}
    auth_result = rucio.gateway.permission.has_permission(issuer=issuer, vo=vo, action='add_did', kwargs=kwargs, session=session)
    if not auth_result.allowed:
        raise AccessDenied('Account %s can not add data identifier to scope %s. %s' % (issuer, scope, auth_result.message))

    owner_account = None if account is None else InternalAccount(account, vo=vo)
    issuer_account = InternalAccount(issuer, vo=vo)
    internal_scope = InternalScope(scope, vo=vo)
    for d in dids:
        d['scope'] = InternalScope(d['scope'], vo=vo)
    for r in rules:
        r['account'] = InternalAccount(r['account'], vo=vo)

    rse_id = None
    if rse is not None:
        rse_id = get_rse_id(rse=rse, vo=vo, session=session)

    if did_type == 'DATASET':
        # naming_convention validation
        extra_meta = naming_convention.validate_name(scope=internal_scope, name=name, did_type='D', session=session)

        # merge extra_meta with meta
        for k in extra_meta or {}:
            if k not in meta:
                meta[k] = extra_meta[k]
            elif meta[k] != extra_meta[k]:
                print("Provided metadata %s doesn't match the naming convention: %s != %s" % (k, meta[k], extra_meta[k]))
                raise InvalidObject("Provided metadata %s doesn't match the naming convention: %s != %s" % (k, meta[k], extra_meta[k]))

        # Validate metadata
        meta_convention_core.validate_meta(meta=meta, did_type=DIDType[did_type.upper()], session=session)

    return did.add_did(scope=internal_scope, name=name, did_type=DIDType[did_type.upper()], account=owner_account or issuer_account,
                       statuses=statuses, meta=meta, rules=rules, lifetime=lifetime,
                       dids=dids, rse_id=rse_id, session=session)


@transactional_session
def add_dids(
    dids: 'Sequence[dict[str, Any]]',
    issuer: str,
    vo: str = 'def',
    *,
    session: "Session"
) -> None:
    """
    Bulk Add did.

    :param dids: A list of dids.
    :param issuer: The issuer account.
    :param vo: The VO to act on.
    :param session: The database session in use.
    """
    for d in dids:
        if 'rse' in d:
            rse_id = None
            if d['rse'] is not None:
                rse_id = get_rse_id(rse=d['rse'], vo=vo, session=session)
            d['rse_id'] = rse_id

    kwargs = {'issuer': issuer, 'dids': dids}
    auth_result = rucio.gateway.permission.has_permission(issuer=issuer, vo=vo, action='add_dids', kwargs=kwargs, session=session)
    if not auth_result.allowed:
        raise AccessDenied('Account %s can not bulk add data identifier. %s' % (issuer, auth_result.message))

    issuer_account = InternalAccount(issuer, vo=vo)
    for d in dids:
        d['scope'] = InternalScope(d['scope'], vo=vo)
        if 'account' in d.keys():
            d['account'] = InternalAccount(d['account'], vo=vo)
        if 'dids' in d.keys():
            for child in d['dids']:
                child['scope'] = InternalScope(child['scope'], vo=vo)
    return did.add_dids(dids, account=issuer_account, session=session)


@transactional_session
def attach_dids(
    scope: str,
    name: str,
    attachment: dict[str, Any],
    issuer: str,
    vo='def',
    *,
    session: "Session"
) -> None:
    """
    Append content to data did.

    :param attachment: The attachment.
    :param issuer: The issuer account.
    :param vo: The VO to act on.
    :param session: The database session in use.
    """
    validate_schema(name='attachment', obj=attachment, vo=vo)

    rse_id = None
    if 'rse' in attachment:
        if attachment['rse'] is not None:
            rse_id = get_rse_id(rse=attachment['rse'], vo=vo, session=session)
        attachment['rse_id'] = rse_id

    kwargs = {'scope': scope, 'name': name, 'attachment': attachment}
    auth_result = rucio.gateway.permission.has_permission(issuer=issuer, vo=vo, action='attach_dids', kwargs=kwargs, session=session)
    if not auth_result.allowed:
        raise AccessDenied('Account %s can not add data identifiers to %s:%s. %s' % (issuer, scope, name, auth_result.message))

    internal_scope = InternalScope(scope, vo=vo)
    issuer_account = InternalAccount(issuer, vo=vo)
    if 'account' in attachment.keys():
        attachment['account'] = InternalAccount(attachment['account'], vo=vo)
    for d in attachment['dids']:
        d['scope'] = InternalScope(d['scope'], vo=vo)
        if 'account' in d.keys():
            d['account'] = InternalAccount(d['account'], vo=vo)

    if rse_id is not None:
        dids = did.attach_dids(scope=internal_scope, name=name, dids=attachment['dids'],
                               account=attachment.get('account', issuer_account), rse_id=rse_id, session=session)
    else:
        dids = did.attach_dids(scope=internal_scope, name=name, dids=attachment['dids'],
                               account=attachment.get('account', issuer_account), session=session)

    return dids


@transactional_session
def attach_dids_to_dids(
    attachments: 'Sequence[dict[str, Any]]',
    issuer: str,
    ignore_duplicate: bool = False,
    vo: str = 'def',
    *,
    session: "Session"
) -> None:
    """
    Append content to dids.

    :param attachments: The contents.
    :param issuer: The issuer account.
    :param ignore_duplicate: If True, ignore duplicate entries.
    :param vo: The VO to act on.
    :param session: The database session in use.
    """
    validate_schema(name='attachments', obj=attachments, vo=vo)

    for a in attachments:
        if 'rse' in a:
            rse_id = None
            if a['rse'] is not None:
                rse_id = get_rse_id(rse=a['rse'], vo=vo, session=session)
            a['rse_id'] = rse_id

    auth_result = rucio.gateway.permission.has_permission(issuer=issuer, vo=vo, action='attach_dids_to_dids', kwargs={'attachments': attachments}, session=session)
    if not auth_result.allowed:
        raise AccessDenied('Account %s can not add data identifiers. %s' % (issuer, auth_result.message))

    issuer_account = InternalAccount(issuer, vo=vo)
    for attachment in attachments:
        attachment['scope'] = InternalScope(attachment['scope'], vo=vo)
        for d in attachment['dids']:
            d['scope'] = InternalScope(d['scope'], vo=vo)
            if 'account' in d.keys():
                d['account'] = InternalAccount(d['account'], vo=vo)

    return did.attach_dids_to_dids(attachments=attachments, account=issuer_account,
                                   ignore_duplicate=ignore_duplicate, session=session)


@transactional_session
def detach_dids(
    scope: str,
    name: str,
    dids: 'Sequence[dict[str, Any]]',
    issuer: str,
    vo: str = 'def',
    *,
    session: "Session"
) -> None:
    """
    Detach data identifier

    :param scope: The scope name.
    :param name: The data identifier name.
    :param dids: The content.
    :param issuer: The issuer account.
    :param vo: The VO to act on.
    :param session: The database session in use.
    """
    kwargs = {'scope': scope, 'name': name, 'dids': dids, 'issuer': issuer}
    auth_result = rucio.gateway.permission.has_permission(issuer=issuer, vo=vo, action='detach_dids', kwargs=kwargs, session=session)
    if not auth_result.allowed:
        raise AccessDenied('Account %s can not detach data identifiers from %s:%s. %s' % (issuer, scope, name, auth_result.message))

    internal_scope = InternalScope(scope, vo=vo)
    for d in dids:
        d['scope'] = InternalScope(d['scope'], vo=vo)

    return did.detach_dids(scope=internal_scope, name=name, dids=dids, session=session)


@stream_session
def list_new_dids(
    did_type: Optional[str] = None,
    thread: Optional[int] = None,
    total_threads: Optional[int] = None,
    chunk_size: int = 1000,
    vo: str = 'def',
    *,
    session: "Session"
) -> 'Iterator[dict[str, Any]]':
    """
    List recent identifiers.

    :param did_type : The DID type.
    :param thread: The assigned thread for this necromancer.
    :param total_threads: The total number of threads of all necromancers.
    :param chunk_size: Number of requests to return per yield.
    :param vo: The VO to act on.
    :param session: The database session in use.
    """
    dids = did.list_new_dids(did_type=did_type and DIDType[did_type.upper()], thread=thread, total_threads=total_threads, chunk_size=chunk_size, session=session)
    for d in dids:
        if d['scope'].vo == vo:
            yield gateway_update_return_dict(d, session=session)


@transactional_session
def set_new_dids(
    dids: 'Sequence[dict[str, Any]]',
    new_flag: bool = True,
    vo: str = 'def',
    *,
    session: "Session"
) -> bool:
    """
    Set/reset the flag new

    :param scope: The scope name.
    :param name: The data identifier name.
    :param new_flag: A boolean to flag new DIDs.
    :param vo: The VO to act on.
    :param session: The database session in use.
    """
    for d in dids:
        d['scope'] = InternalScope(d['scope'], vo=vo)

    return did.set_new_dids(dids, new_flag, session=session)


@stream_session
def list_content(
    scope: str,
    name: str,
    vo: str = 'def',
    *,
    session: "Session"
) -> 'Iterator[dict[str, Any]]':
    """
    List data identifier contents.

    :param scope: The scope name.
    :param name: The data identifier name.
    :param vo: The VO to act on.
    :param session: The database session in use.
    """

    internal_scope = InternalScope(scope, vo=vo)

    dids = did.list_content(scope=internal_scope, name=name, session=session)
    for d in dids:
        yield gateway_update_return_dict(d, session=session)


@stream_session
def list_content_history(
    scope: str,
    name: str,
    vo='def',
    *,
    session: "Session"
) -> 'Iterator[dict[str, Any]]':
    """
    List data identifier contents history.

    :param scope: The scope name.
    :param name: The data identifier name.
    :param vo: The VO to act on.
    :param session: The database session in use.
    """

    internal_scope = InternalScope(scope, vo=vo)

    dids = did.list_content_history(scope=internal_scope, name=name, session=session)

    for d in dids:
        yield gateway_update_return_dict(d, session=session)


@stream_session
def bulk_list_files(
    dids: 'Iterable[dict[str, Any]]',
    long: bool = False,
    vo: str = 'def',
    *,
    session: "Session"
) -> 'Iterator[dict[str, Any]]':
    """
    List file contents of a list of data identifiers.

    :param dids:       A list of DIDs.
    :param long:       A boolean to choose if more metadata are returned or not.
    :param vo:         The VO to act on.
    :param session:    The database session in use.
    """

    for did_ in dids:
        did_['scope'] = InternalScope(did_['scope'], vo=vo)

    for file_ in did.bulk_list_files(dids=dids, long=long, session=session):
        yield gateway_update_return_dict(file_, session=session)


@stream_session
def list_files(
    scope: str,
    name: str,
    long: bool,
    vo: str = 'def',
    *,
    session: "Session"
) -> 'Iterator[dict[str, Any]]':
    """
    List data identifier file contents.

    :param scope: The scope name.
    :param name: The data identifier name.
    :param long:       A boolean to choose if GUID is returned or not.
    :param vo: The VO to act on.
    :param session: The database session in use.
    """

    internal_scope = InternalScope(scope, vo=vo)

    dids = did.list_files(scope=internal_scope, name=name, long=long, session=session)

    for d in dids:
        yield gateway_update_return_dict(d, session=session)


@stream_session
def scope_list(
    scope: str,
    name: Optional[str] = None,
    recursive: bool = False,
    vo: str = 'def',
    *,
    session: "Session"
) -> 'Iterator[dict[str, Any]]':
    """
    List data identifiers in a scope.

    :param scope: The scope name.
    :param name: The data identifier name.
    :param recursive: boolean, True or False.
    :param vo: The VO to act on.
    :param session: The database session in use.
    """

    internal_scope = InternalScope(scope, vo=vo)

    dids = did.scope_list(internal_scope, name=name, recursive=recursive, session=session)

    for d in dids:
        ret_did = deepcopy(d)
        ret_did['scope'] = ret_did['scope'].external
        if ret_did['parent'] is not None:
            ret_did['parent']['scope'] = ret_did['parent']['scope'].external
        yield ret_did


@read_session
def get_did(scope: str, name: str, dynamic_depth: Optional[DIDType] = None, vo: str = 'def', *, session: "Session") -> "dict[str, Any]":
    """
    Retrieve a single data did.

    :param scope: The scope name.
    :param name: The data identifier name.
    :param dynamic_depth: the DID type to use as source for estimation of this DIDs length/bytes.
    If set to None, or to a value which doesn't make sense (ex: requesting depth = CONTAINER for a did of type DATASET)
    will not compute the size dynamically.
    :param vo: The VO to act on.
    :return did: Dictionary containing {'name', 'scope', 'type'}, Exception otherwise
    :param session: The database session in use.
    """

    internal_scope = InternalScope(scope, vo=vo)

    d = did.get_did(scope=internal_scope, name=name, dynamic_depth=dynamic_depth, session=session)
    return gateway_update_return_dict(d, session=session)


@transactional_session
def set_metadata(
    scope: str,
    name: str,
    key: str,
    value: Any,
    issuer: str,
    recursive: bool = False,
    vo: str = 'def',
    *,
    session: "Session"
) -> None:
    """
    Add metadata to data did.

    :param scope: The scope name.
    :param name: The data identifier name.
    :param key: the key.
    :param value: the value.
    :param issuer: The issuer account.
    :param recursive: Option to propagate the metadata update to content.
    :param vo: The VO to act on.
    :param session: The database session in use.
    """
    kwargs = {'scope': scope, 'name': name, 'key': key, 'value': value, 'issuer': issuer}

    if key in RESERVED_KEYS:
        raise AccessDenied('Account %s can not change this metadata value to data identifier %s:%s' % (issuer, scope, name))

    auth_result = rucio.gateway.permission.has_permission(issuer=issuer, vo=vo, action='set_metadata', kwargs=kwargs, session=session)
    if not auth_result.allowed:
        raise AccessDenied('Account %s can not add metadata to data identifier %s:%s. %s' % (issuer, scope, name, auth_result.message))

    internal_scope = InternalScope(scope, vo=vo)
    return did.set_metadata(scope=internal_scope, name=name, key=key, value=value, recursive=recursive, session=session)


@transactional_session
def set_metadata_bulk(
    scope: str,
    name: str,
    meta: dict[str, Any],
    issuer: str,
    recursive: bool = False,
    vo: str = 'def',
    *,
    session: "Session"
) -> None:
    """
    Add metadata to data did.

    :param scope: The scope name.
    :param name: The data identifier name.
    :param meta: the key-values.
    :param issuer: The issuer account.
    :param recursive: Option to propagate the metadata update to content.
    :param vo: The VO to act on.
    :param session: The database session in use.
    """
    kwargs = {'scope': scope, 'name': name, 'meta': meta, 'issuer': issuer}

    for key in meta:
        if key in RESERVED_KEYS:
            raise AccessDenied('Account %s can not change the value of the metadata key %s to data identifier %s:%s' % (issuer, key, scope, name))

    auth_result = rucio.gateway.permission.has_permission(issuer=issuer, vo=vo, action='set_metadata_bulk', kwargs=kwargs, session=session)
    if not auth_result.allowed:
        raise AccessDenied('Account %s can not add metadata to data identifier %s:%s. %s' % (issuer, scope, name, auth_result.message))

    internal_scope = InternalScope(scope, vo=vo)
    return did.set_metadata_bulk(scope=internal_scope, name=name, meta=meta, recursive=recursive, session=session)


@transactional_session
def set_dids_metadata_bulk(
    dids: 'Iterable[dict[str, Any]]',
    issuer: str,
    recursive: bool = False,
    vo: str = 'def',
    *,
    session: "Session"
) -> None:
    """
    Add metadata to a list of data identifiers.

    :param issuer: The issuer account.
    :param dids: A list of dids including metadata.
    :param recursive: Option to propagate the metadata update to content.
    :param vo: The VO to act on.
    :param session: The database session in use.
    """

    for entry in dids:
        kwargs = {'scope': entry['scope'], 'name': entry['name'], 'meta': entry['meta'], 'issuer': issuer}
        auth_result = rucio.gateway.permission.has_permission(issuer=issuer, vo=vo, action='set_metadata_bulk', kwargs=kwargs, session=session)
        if not auth_result.allowed:
            raise AccessDenied('Account %s can not add metadata to data identifier %s:%s. %s' % (issuer, entry['scope'], entry['name'], auth_result.message))
        entry['scope'] = InternalScope(entry['scope'], vo=vo)
        meta = entry['meta']
        for key in meta:
            if key in RESERVED_KEYS:
                raise AccessDenied('Account %s can not change the value of the metadata key %s to data identifier %s:%s' % (issuer, key, entry['scope'], entry['name']))

    return did.set_dids_metadata_bulk(dids=dids, recursive=recursive, session=session)


@read_session
def get_metadata(
    scope: str,
    name: str,
    plugin: str = 'DID_COLUMN',
    vo: str = 'def',
    *,
    session: "Session"
) -> dict[str, Any]:
    """
    Get data identifier metadata

    :param scope: The scope name.
    :param name: The data identifier name.
    :param vo: The VO to act on.
    :param plugin: The metadata plugin to query, 'ALL' for all available plugins
    :param session: The database session in use.
    """

    internal_scope = InternalScope(scope, vo=vo)

    d = did.get_metadata(scope=internal_scope, name=name, plugin=plugin, session=session)
    return gateway_update_return_dict(d, session=session)


@stream_session
def get_metadata_bulk(
    dids: 'Iterable[dict[str, Any]]',
    inherit: bool = False,
    plugin: str = 'DID_COLUMN',
    vo: str = 'def',
    *,
    session: "Session"
) -> 'Iterator[dict[str, Any]]':
    """
    Get metadata for a list of dids
    :param dids:               A list of dids.
    :param inherit:            A boolean. If set to true, the metadata of the parent are concatenated.
    :param plugin:             The metadata plugin to query, 'ALL' for all available plugins
    :param vo:                 The VO to act on.
    :param session: The database session in use.
    """

    validate_schema(name='dids', obj=dids, vo=vo)
    for entry in dids:
        entry['scope'] = InternalScope(entry['scope'], vo=vo)
    meta = did.get_metadata_bulk(dids, inherit=inherit, plugin=plugin, session=session)
    for met in meta:
        yield gateway_update_return_dict(met, session=session)


@transactional_session
def delete_metadata(
    scope: str,
    name: str,
    key: str,
    vo: str = 'def',
    *,
    session: "Session"
) -> None:
    """
    Delete a key from the metadata column

    :param scope: the scope of did
    :param name: the name of the did
    :param key: the key to be deleted
    :param vo: The VO to act on.
    :param session: The database session in use.
    """

    internal_scope = InternalScope(scope, vo=vo)
    return did.delete_metadata(scope=internal_scope, name=name, key=key, session=session)


@transactional_session
def set_status(
    scope: str,
    name: str,
    issuer: str,
    vo: str = 'def',
    *,
    session: "Session",
    **kwargs
) -> None:
    """
    Set data identifier status

    :param scope: The scope name.
    :param name: The data identifier name.
    :param issuer: The issuer account.
    :param kwargs:  Keyword arguments of the form status_name=value.
    :param vo: The VO to act on.
    :param session: The database session in use.
    """

    auth_result = rucio.gateway.permission.has_permission(issuer=issuer, vo=vo, action='set_status', kwargs={'scope': scope, 'name': name, 'issuer': issuer}, session=session)
    if not auth_result.allowed:
        raise AccessDenied('Account %s can not set status on data identifier %s:%s. %s' % (issuer, scope, name, auth_result.message))

    internal_scope = InternalScope(scope, vo=vo)

    return did.set_status(scope=internal_scope, name=name, session=session, **kwargs)


@stream_session
def get_dataset_by_guid(
    guid: str,
    vo: str = 'def',
    *,
    session: "Session"
) -> 'Iterator[dict[str, Any]]':
    """
    Get the parent datasets for a given GUID.
    :param guid: The GUID.
    :param vo: The VO to act on.
    :param session: The database session in use.

    :returns: A did
    """
    dids = did.get_dataset_by_guid(guid=guid, session=session)

    for d in dids:
        if d['scope'].vo != vo:
            raise RucioException('GUID unavailable on VO {}'.format(vo))
        yield gateway_update_return_dict(d, session=session)


@stream_session
def list_parent_dids(
    scope: str,
    name: str,
    vo: str = 'def',
    *,
    session: "Session"
) -> 'Iterator[dict[str, Any]]':
    """
    List parent datasets and containers of a did.

    :param scope:   The scope.
    :param name:    The name.
    :param vo:      The VO to act on.
    :param session: The database session in use.
    """

    internal_scope = InternalScope(scope, vo=vo)

    dids = did.list_parent_dids(scope=internal_scope, name=name, session=session)

    for d in dids:
        yield gateway_update_return_dict(d, session=session)


@transactional_session
def create_did_sample(
    input_scope: str,
    input_name: str,
    output_scope: str,
    output_name: str,
    issuer: str,
    nbfiles: str,
    vo: str = 'def',
    *,
    session: "Session"
):
    """
    Create a sample from an input collection.

    :param input_scope: The scope of the input DID.
    :param input_name: The name of the input DID.
    :param output_scope: The scope of the output dataset.
    :param output_name: The name of the output dataset.
    :param account: The account.
    :param nbfiles: The number of files to register in the output dataset.
    :param issuer: The issuer account.
    :param vo: The VO to act on.
    :param session: The database session in use.
    """
    kwargs = {'issuer': issuer, 'scope': output_scope}
    auth_result = rucio.gateway.permission.has_permission(issuer=issuer, vo=vo, action='create_did_sample', kwargs=kwargs, session=session)
    if not auth_result.allowed:
        raise AccessDenied('Account %s can not bulk add data identifier. %s' % (issuer, auth_result.message))

    input_internal_scope = InternalScope(input_scope, vo=vo)
    output_internal_scope = InternalScope(output_scope, vo=vo)

    issuer_account = InternalAccount(issuer, vo=vo)

    return did.create_did_sample(input_scope=input_internal_scope, input_name=input_name, output_scope=output_internal_scope, output_name=output_name,
                                 account=issuer_account, nbfiles=nbfiles, session=session)


@transactional_session
def resurrect(
    dids: 'Iterable[dict[str, Any]]',
    issuer: str,
    vo: str = 'def',
    *,
    session: "Session"
) -> None:
    """
    Resurrect DIDs.

    :param dids: A list of dids.
    :param issuer: The issuer account.
    :param vo: The VO to act on.
    :param session: The database session in use.
    """
    kwargs = {'issuer': issuer}
    auth_result = rucio.gateway.permission.has_permission(issuer=issuer, vo=vo, action='resurrect', kwargs=kwargs, session=session)
    if not auth_result.allowed:
        raise AccessDenied('Account %s can not resurrect data identifiers. %s' % (issuer, auth_result.message))
    validate_schema(name='dids', obj=dids, vo=vo)

    for d in dids:
        d['scope'] = InternalScope(d['scope'], vo=vo)

    return did.resurrect(dids=dids, session=session)


@stream_session
def list_archive_content(
    scope: str,
    name: str,
    vo: str = 'def',
    *,
    session: "Session"
) -> 'Iterator[dict[str, Any]]':
    """
    List archive contents.

    :param scope: The archive scope name.
    :param name: The archive data identifier name.
    :param vo: The VO to act on.
    :param session: The database session in use.
    """

    internal_scope = InternalScope(scope, vo=vo)

    dids = did.list_archive_content(scope=internal_scope, name=name, session=session)
    for d in dids:
        yield gateway_update_return_dict(d, session=session)


@transactional_session
def add_did_to_followed(
    scope: str,
    name: str,
    account: str,
    *,
    session: "Session",
    vo: str = 'def'
) -> None:
    """
    Mark a did as followed by the given account

    :param scope: The scope name.
    :param name: The data identifier name.
    :param account: The account owner.
    :param session: The database session in use.
    """
    internal_scope = InternalScope(scope, vo=vo)
    internal_account = InternalAccount(account, vo=vo)
    return did.add_did_to_followed(scope=internal_scope, name=name, account=internal_account, session=session)


@transactional_session
def add_dids_to_followed(
    dids: 'Iterable[Mapping[str, Any]]',
    account: str,
    *,
    session: "Session",
    vo: str = 'def'
) -> None:
    """
    Bulk mark datasets as followed

    :param dids: A list of dids.
    :param account: The account owner.
    :param session: The database session in use.
    """
    internal_account = InternalAccount(account, vo=vo)
    return did.add_dids_to_followed(dids=dids, account=internal_account, session=session)


@stream_session
def get_users_following_did(
    name: str,
    scope: str,
    *,
    session: "Session",
    vo: str = 'def'
) -> 'Iterator[dict[str, str]]':
    """
    Return list of users following a did

    :param scope: The scope name.
    :param name: The data identifier name.
    :param session: The database session in use.
    """
    internal_scope = InternalScope(scope, vo=vo)
    users = did.get_users_following_did(name=name, scope=internal_scope, session=session)
    for user in users:
        user['user'] = user['user'].external
        yield user


@transactional_session
def remove_did_from_followed(
    scope: str,
    name: str,
    account: str,
    issuer: str,
    *,
    session: "Session",
    vo: str = 'def'
) -> None:
    """
    Mark a did as not followed

    :param scope: The scope name.
    :param name: The data identifier name.
    :param account: The account owner.
    :param session: The database session in use.
    :param issuer: The issuer account
    """
    kwargs = {'scope': scope, 'issuer': issuer}
    auth_result = rucio.gateway.permission.has_permission(issuer=issuer, vo=vo, action='remove_did_from_followed', kwargs=kwargs, session=session)
    if not auth_result.allowed:
        raise AccessDenied('Account %s can not remove data identifiers from followed table. %s' % (issuer, auth_result.message))

    internal_scope = InternalScope(scope, vo=vo)
    internal_account = InternalAccount(account, vo=vo)
    return did.remove_did_from_followed(scope=internal_scope, name=name, account=internal_account, session=session)


@transactional_session
def remove_dids_from_followed(
    dids: 'Iterable[Mapping[str, Any]]',
    account: str,
    issuer: str,
    *,
    session: "Session",
    vo: str = 'def'
) -> None:
    """
    Bulk mark datasets as not followed

    :param dids: A list of dids.
    :param account: The account owner.
    :param session: The database session in use.
    """
    kwargs = {'dids': dids, 'issuer': issuer}
    auth_result = rucio.gateway.permission.has_permission(issuer=issuer, vo=vo, action='remove_dids_from_followed', kwargs=kwargs, session=session)
    if not auth_result.allowed:
        raise AccessDenied('Account %s can not bulk remove data identifiers from followed table. %s' % (issuer, auth_result.message))

    internal_account = InternalAccount(account, vo=vo)
    return did.remove_dids_from_followed(dids=dids, account=internal_account, session=session)



# Copyright European Organization for Nuclear Research (CERN) since 2012
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import click

from rucio.cli.bin_legacy.rucio import list_dataset_replicas, list_file_replicas, list_suspicious_replicas
from rucio.cli.bin_legacy.rucio_admin import declare_bad_file_replicas, declare_temporary_unavailable_replicas, quarantine_replicas, set_tombstone
from rucio.cli.utils import Arguments


@click.group()
def replica():
    """Manage replicas - DIDs with locations on RSEs"""


@replica.group("list")
def replica_list():
    """List replicas (file or collection-types)"""


@replica_list.command("file")
@click.argument("dids", nargs=-1)
@click.option("--protocols", help="Protocol used to access a replicas (i.e. https, root, srm)")
@click.option(
    "--all-states",
    help="To select all replicas (including unavailable ones).\
                Also gets information about the current state of a DID in each RSE",
    is_flag=True,
    default=False,
)
@click.option("--pfns", is_flag=True, help="Show only the PFNs", default=False)
@click.option("--domain", default="all", type=click.Choice(["wan", "lan", "all"]), help="Force the networking domain")
@click.option(
    "--link",
    help="Symlink PFNs with directory substitution.\
                For example: rucio list-file-replicas --rse RSE_TEST --link /eos/:/eos/ scope:datasetname",
)
@click.option("--missing", is_flag=True, default=False, help="To list missing replicas at a RSE-Expression. Must be used with --rses option")
@click.option("--metalink", is_flag=True, default=False, help="Output available replicas as metalink")
@click.option("--no-resolve-archives", is_flag=True, default=False, help="Do not resolve archives which may contain the files", required=False)
@click.option("--sort", help="Replica sort algorithm. Available options: geoip (default), random")
@click.option("--rses", "--rse-exp", "rses", help="The RSE filter expression")
@click.option("--human", default=True, hidden=True)
@click.pass_context
def list_(ctx, dids, protocols, all_states, pfns, domain, link, missing, metalink, no_resolve_archives, sort, rses, human):
    """List the replicas of a DID and its PFNs. By default all states, even unavailable, are shown"""
    args = {"dids": dids, "protocols": protocols, "all_states": all_states, "pfns": pfns, "domain": domain, "link": link, "missing": missing, "metalink": metalink, "no_resolve_archives": no_resolve_archives, "sort": sort, "rses": rses, "human": human}
    list_file_replicas(Arguments(args), ctx.obj.client, ctx.obj.logger, ctx.obj.console, ctx.obj.spinner)


@replica_list.command("dataset")
@click.argument("dids", nargs=-1)
@click.option("--deep", default=False, is_flag=True, help="Make a deep check, checking the contents of datasets in datasets")
@click.option("--csv", help="Write output to comma separated values", is_flag=True, default=False)
@click.pass_context
def list_dataset(ctx, dids, deep, csv):
    """List dataset replicas"""
    args = Arguments({"dids": dids, "deep": deep, "csv": csv})
    list_dataset_replicas(args, ctx.obj.client, ctx.obj.logger, ctx.obj.console, ctx.obj.spinner)


@replica.command("remove")
@click.argument("dids", nargs=-1)
@click.option("--rse", "--rse-name", "rse", required=True)
@click.pass_context
def remove(ctx, dids, rse):
    "Set a replica for removal by adding a tombstone which will mark the replica as ready for deletion by a reaper daemon"
    # TODO: Fix set_tombstone to not expect a comma separated DID str
    dids = ",".join(dids)
    set_tombstone(Arguments({"dids": dids, "rse": rse}), ctx.obj.client, ctx.obj.logger, ctx.obj.console, ctx.obj.spinner)


@replica.group()
@click.help_option("-h", "--help")
def state():
    """Manage the state of replicas"""


@state.command("list")
@click.argument("state-type", type=click.Choice(["suspicious"]))
@click.option("--rses", "--rse-exp", help="RSE name or expression")  # TODO remap rse_expression to rses (for consistency)
@click.option("--younger-than", help='List files that have been marked suspicious since the date "younger_than", e.g. 2021-11-29T00:00:00')  # NOQA: E501
@click.option("--n-attempts", help="Minimum number of failed attempts to access a suspicious file")
@click.pass_context
def state_list(ctx, state_type, rses, younger_than, n_attempts):
    """List replicas by state. WARNING: Only implemented for 'suspicious'"""

    if state_type != "suspicious":
        raise ValueError(f"Cannot list state by {state_type}, please choose from ('suspicious')")
    list_suspicious_replicas(Arguments({"rse_expression": rses, "younger_than": younger_than, "nattempts": n_attempts}), ctx.obj.client, ctx.obj.logger, ctx.obj.console, ctx.obj.spinner)


@state.group("update")
@click.help_option("-h", "--help")
def state_update():
    "Change the state of replicas"


@state_update.command("bad")
@click.argument("replicas", nargs=-1)
@click.option("--reason", required=True, help="Reason")
@click.option("--as-file", is_flag=True, default=False, help="[REPLICAS] arg is a path to a file of replicas to update")
@click.option("--collection", is_flag=True, default=False, help="Items in the collection DID are also marked as bad")
@click.option("--lfn", is_flag=True, default=False, help="[REPLICAS] arg is a path to a file of LFNs. Requires --rse and --scope")
@click.option("--scope", help="Common scope for bad replicas specified with LFN list, ignored without --lfn")
@click.option("--rse", "--rse-name", help="Common RSE for bad replicas specified with LFN list, ignored without --lfn")
@click.pass_context
def update_bad(ctx, replicas, reason, as_file, collection, lfn, scope, rse):
    """Mark a replica bad"""
    args = {"reason": reason, "allow_collection": collection, "scope": scope, "rse": rse}
    if as_file:
        args["inputfile"] = replicas
    elif lfn:
        if (scope is None) or (rse is None):
            raise ValueError("Scope and RSE are required when using LFNs")
        args["lfns"] = replicas
    else:
        args["listbadfiles"] = replicas
    declare_bad_file_replicas(Arguments(args), ctx.obj.client, ctx.obj.logger, ctx.obj.console, ctx.obj.spinner)


@state_update.command("unavailable")
@click.argument("replicas", nargs=-1)
@click.option("--reason", required=True, help="Reason")
@click.option("--as-file", is_flag=True, default=False, help="[REPLICAS] arg is a path to a file of names to update")
@click.option("--duration", required=True, type=int, help="Timeout (in seconds) after which the replicas will become available again")
@click.pass_context
def update_unavailable(ctx, replicas, reason, as_file, duration):
    """Declare a replica unavailable"""
    args = {"reason": reason, "duration": duration}
    if as_file:
        args["inputfile"] = replicas
    else:
        args["listbadfiles"] = replicas
    declare_temporary_unavailable_replicas(Arguments(args), ctx.obj.client, ctx.obj.logger, ctx.obj.console, ctx.obj.spinner)


@state_update.command("quarantine")
@click.argument("replicas", nargs=-1)
@click.option("--as-file", is_flag=True, default=False, help="[REPLICAS] arg is a path to a file of names to update")
@click.option("--rse", "--rse-name")  # TODO What does this do?
@click.pass_context
def update_quarantine(ctx, replicas, as_file, rse):
    """Quarantine a replica"""
    args = {"rse": rse}
    if as_file:
        args["paths_file"] = replicas
    else:
        args["paths_list"] = replicas

    # TODO Add a reason option
    quarantine_replicas(Arguments(args), ctx.obj.client, ctx.obj.logger, ctx.obj.console, ctx.obj.spinner)



# Copyright European Organization for Nuclear Research (CERN) since 2012
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from json import dumps
from typing import TYPE_CHECKING

from flask import Flask, Response, request

from rucio.gateway.did import list_archive_content
from rucio.web.rest.flaskapi.authenticated_bp import AuthenticatedBlueprint
from rucio.web.rest.flaskapi.v1.common import ErrorHandlingMethodView, check_accept_header_wrapper_flask, generate_http_error_flask, parse_scope_name, response_headers, try_stream

if TYPE_CHECKING:
    from collections.abc import Iterator


class Archive(ErrorHandlingMethodView):
    """ REST APIs for archive. """

    @check_accept_header_wrapper_flask(['application/x-json-stream'])
    def get(self, scope_name: str) -> Response:
        """
        ---
        summary: List
        description: List archive contents.
        tags:
          - Archive
        parameters:
        - name: scope_name
          in: path
          description: The data identifier of the scope.
          schema:
            type: string
          style: simple
        responses:
          201:
            description: OK
            content:
              application/x-json-stream:
                schema:
                  type: array
                  items:
                    type: object
                    properties:
                      scope:
                        description: The scope of the did.
                        type: string
                      name:
                        description: The name of the did.
                        type: string
                      bytes:
                        description: The number of bytes.
                        type: integer
                      adler32:
                        description: The adler32 checksum.
                        type: string
                      md5:
                        description: The md5 checksum.
                        type: string
          400:
            description: Invalid value
          406:
            description: Not acceptable
        """
        try:
            scope, name = parse_scope_name(scope_name, request.environ.get('vo'))

            def generate(vo: str) -> 'Iterator[str]':
                for file in list_archive_content(scope=scope, name=name, vo=vo):
                    yield dumps(file) + '\n'

            return try_stream(generate(vo=request.environ.get('vo', 'def')))
        except ValueError as error:
            return generate_http_error_flask(400, error)


def blueprint() -> AuthenticatedBlueprint:
    bp = AuthenticatedBlueprint('archives', __name__, url_prefix='/archives')

    archive_view = Archive.as_view('archive')
    bp.add_url_rule('/<path:scope_name>/files', view_func=archive_view, methods=['get', ])

    bp.after_request(response_headers)
    return bp


def make_doc() -> Flask:
    """ Only used for sphinx documentation """
    doc_app = Flask(__name__)
    doc_app.register_blueprint(blueprint())
    return doc_app
